using System;
using JJC.Psharp.Lang;
using JJC.Psharp.Predicates;
using System.Xml;
using System.Collections;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Reflection;

namespace DCAServer {
	/// <summary>
	/// </summary>
	
	class PredicateDescriptor {
		public string name;
		public ArrayList key;

		public PredicateDescriptor( string name, ArrayList key ) {
			this.name = name;
			this.key = key;
		}
	}

	class DCAServer {

		ArrayList predicates;

		static int MAX_AGENTS = 10;

		AgentConnection[] agents = new AgentConnection[ MAX_AGENTS ];

		public PushbackReader GetStringReader( string s ) {
			StringReader sr = new StringReader( s );
			return new PushbackReader( (TextReader)sr ); 
		}

		private void ReadTable( XmlTextReader reader ) {
			while (reader.Read()) {
				switch (reader.NodeType) {
					case XmlNodeType.Element:
						if( reader.Name.Equals( "predicate" ) ) {
							ReadPredicateData( reader, reader.GetAttribute( "name" ) );
						} else
							throw new Exception( );
						break;
					case XmlNodeType.EndElement:
						break;
					case XmlNodeType.Whitespace:
						break;
					default:
						throw new Exception( );
				}       
			}
		}

		private void ReadPredicateData( XmlTextReader reader, string name ) {
			ArrayList key = new ArrayList( );
			while( reader.Read( ) ) {
				switch( reader.NodeType ) {
					case XmlNodeType.Element:
						if( reader.Name.Equals( "field" ) ) { 
							string isKey = reader.GetAttribute( "key" );
							key.Add( isKey );
							reader.Read( );

							if( !( reader.Name.Equals( "field" ) ) ||
								!( reader.NodeType == XmlNodeType.EndElement ) )
								throw new Exception( );
						}
						else
							throw new Exception( );
						break;
					case XmlNodeType.EndElement:
						if( !reader.Name.Equals( "predicate" ) )
							throw new Exception( );
						else {
							predicates.Add( new PredicateDescriptor( name, key ) );
							return;
						}
					case XmlNodeType.Whitespace:
						break;
				    default:
						throw new Exception( );
				}       
			}
		}

		private void ReadXMLFile( string file ) {
			XmlTextReader reader = new XmlTextReader( file );

			reader.WhitespaceHandling = System.Xml.WhitespaceHandling.Significant;

			try {
				while (reader.Read()) {
					switch (reader.NodeType) {
						case XmlNodeType.Element:
							if( reader.Name.Equals( "dcatable" ) )
								ReadTable( reader );
							else
								throw new Exception( );
							break;
						case XmlNodeType.EndElement:
							break;
						case XmlNodeType.Whitespace:
							break;
						default:
							throw new Exception( );
					}       
				}           
			} finally {
				if( reader != null )
					reader.Close();
			}
		}

		public void InstanceMain( string[] args ) {

			for( int i = 0; i < MAX_AGENTS; i++ )
				agents[ i ] = null;

			predicates = new ArrayList( );
			
			// args[0] should be a file, we will load the
			// schema for the data therefrom (maybe this should be XML).
			if( args.Length != 1 ) {
				Console.WriteLine( "Usage: DCAServer <data schema file>" );
				return;
			}
			string file = args[ 0 ];
			ReadXMLFile( file );

			// print out the parsed data
			foreach( PredicateDescriptor predicate in predicates ) {
				Console.Write( predicate.name + "( " );
				foreach( string k in predicate.key )
					if( k.Equals( "true" ) )
						Console.Write( "key " );
					else
						Console.Write( "nonkey " );
				Console.WriteLine( ")" );
			}
			Console.WriteLine( );

			// now dump all this data into the prolog engine,
			// so that it can assert relevant predicates.

			PrologInterface sharp = new PrologInterface( );
			sharp.AddAssembly( System.Reflection.Assembly.GetExecutingAssembly() );
			sharp.SetPredicate( new JJC.Psharp.Predicates.DcaInit_1( new CsObjectTerm( this ),  
				new ReturnCs( sharp ) ) );
			sharp.Call( );
		
			SymbolTerm key = SymbolTerm.MakeSymbol( "key", 0 );
			SymbolTerm nonkey = SymbolTerm.MakeSymbol( "nonkey", 0 );
			
			for( int i = 0; i < predicates.Count; i++ ) {

				PredicateDescriptor pd = (PredicateDescriptor)
					(predicates[ i ] );

				string s = pd.name;
				
				ArrayList k = pd.key;
				Term lt = SymbolTerm.MakeSymbol( "[]", 0 );
				for( int j = k.Count - 1; j >= 0; j-- ) {
					if( ((string)k[ j ]).Equals( "true" ) )
						lt = new ListTerm( key, lt );
					else
						lt = new ListTerm( nonkey, lt );
				}
				sharp.SetPredicate( new DcaAssertPred_3(
					SymbolTerm.MakeSymbol( s ),
					new IntegerTerm( k.Count ),
					lt, new ReturnCs( sharp ) ) );
				sharp.Call( );
			}
			RunServer( );
		}

		[STAThread]
		static void Main(string[] args) {
			new DCAServer( ).InstanceMain( args );
		}

		public void Broadcast( string tb ) {
			for( int i = 0; i < MAX_AGENTS; i++ )
				if( agents[i] != null )
					agents[i].Enqueue( tb );
		}

		public void ResetAllPorts( object[] lt ) {
			ResetAgents( lt );
		}

		public void ResetAgents( object[] lt ) {
			for( int i = 0; i < MAX_AGENTS; i++ ) {
				if( agents[i] != null ) {
					foreach( object o in lt ) {
						string s = o.ToString( );
						if( s.EndsWith( "." ) )
							s = s.Substring( 0, s.Length - 1 );
						agents[i].Enqueue( "broadcast_assert(" + s + ")." );
					}
					agents[i].Enqueue( "end." );
				}

			}
		}

		public void BroadcastAssert( string tba ) {
			if( tba.EndsWith( "." ) )
				tba = tba.Substring( 0, tba.Length - 1 );
			Broadcast( "broadcast_assert(" + tba + ")." );
		}

		public void BroadcastRetract( string tbr ) {
			if( tbr.EndsWith( "." ) )
				tbr = tbr.Substring( 0, tbr.Length - 1 );
			Broadcast( "broadcast_retract(" + tbr + ")." );
		}

		public void BroadcastAllRetract( ) {
			Broadcast( "all_retract." );
		}

		public object databaseLock = new object( );

		public void DoSynch( BinaryReader reader, 
									BinaryWriter writer,
									string synchagent ) {
			try {
				int agentNumber = Int32.Parse( synchagent );
				agents[ agentNumber ].SendQueue( writer );
			} catch( Exception e ) {
				Console.WriteLine( "Synchronization failed.\n" + e );
				return;
			}
		}

		public void DoAssert( string tba ) {
			lock( databaseLock ) {
				PrologInterface sharp = new PrologInterface( );
				sharp.SetPredicate( new DcaAssert_1( 
					SymbolTerm.MakeSymbol( tba ), new ReturnCs( sharp ) ) );
				sharp.Call( );
			}
		}

		public void DoRetract( string tbr ) {
			lock( databaseLock ) {
				PrologInterface sharp = new PrologInterface( );
				sharp.SetPredicate( new DcaRetract_1( 
					SymbolTerm.MakeSymbol( tbr ), new ReturnCs( sharp ) ) );
				sharp.Call( );
			}
		}

		public string Strip( string sin ) {
			int i = sin.IndexOf( "(" );
			int j = sin.LastIndexOf( ")" );
			string sout = sin.Substring( i + 1, j - i - 1 );
			return sout;
		}

		public string StripWithDot( string sin ) {
			return Strip( sin ) + ".";
		}

		public object resolveConflictLock = new object( );

		public int ResolveConflict( object[] lt,
			object r, object w ) {
			
			lock( resolveConflictLock ) {
				BinaryReader reader = (BinaryReader)r;
				BinaryWriter writer = (BinaryWriter)w;
				writer.Write( "conflicts." );
				foreach( object o in lt ) 
					writer.Write( o.ToString() + "." );
				writer.Write( "!conflicts." );
				string s = reader.ReadString( );
				return Int32.Parse( s );
			} 
		}

		private void HandleConnect( BinaryReader reader,
			BinaryWriter writer, NetworkStream socketStream,
			Socket connection ) {

			PrologInterface sharp = new PrologInterface( );
			ArrayList agentClauses = new ArrayList( );
			string next;

			for( int i = 0; i < MAX_AGENTS + 1; i++ ) {
				if( i == MAX_AGENTS ) {
					writer.Write( "too_many_agents." );
					return;
				}
				if( agents[ i ] == null ) {
					Console.WriteLine( "Allocated new agent to slot " + i );
					writer.Write( "id( " + i + " )." );
					agents[ i ] = new AgentConnection( i );
					break;
				}
			}

			while( !( next = reader.ReadString( ) ).Equals( "!retractions" ) ) {
				Console.Out.WriteLine( "Received: " + next );
				DoRetract( StripWithDot( next ) );	
			}

			while( !( next = reader.ReadString( ) ).Equals( "!claims." ) ) {
				Console.Out.WriteLine( "Received: " + next );
				agentClauses.Add( next );
				sharp.SetPredicate( new DcaAssertNoCopy_1( 
					SymbolTerm.MakeSymbol( next ), new ReturnCs( sharp ) ) );
				sharp.Call( );
			}
	
			Console.Out.WriteLine( "Now resolving..." );
			sharp.SetPredicate( new ResolveAll_2( 
				new CsObjectTerm( reader ),
				new CsObjectTerm( writer ), 
				new ReturnCs( sharp ) ) );
			sharp.Call( );

			writer.Write( "ok." );
			Console.Out.WriteLine( "ok." );
				
			try { reader.Close( ); } catch { }
			try { writer.Close( ); } catch { }
			try { socketStream.Close( ); } catch { }
			try { connection.Close( ); } catch { }

			Console.Out.WriteLine( "Now resetting all..." );
			sharp.SetPredicate( new ResetAllAgents_0( new ReturnCs( sharp ) ) );
			sharp.Call( );
		}

		private void RunServer( ) {
			TcpListener listener = new TcpListener( 6000 );
			Socket connection;
			NetworkStream socketStream;
			BinaryWriter writer;
			BinaryReader reader;
			string msg;
			try {
				//listener = new TcpListener( 6000 );
				listener.Start( );
				while( true ) {	
					connection = listener.AcceptSocket( );
					socketStream = new NetworkStream( connection );
					writer = new BinaryWriter( socketStream );
					reader = new BinaryReader( socketStream );
					try {
						msg = reader.ReadString( );
						Console.WriteLine( "Message received: " + msg );	
						if( msg.StartsWith( "assert" ) ) {
							DoAssert( StripWithDot( msg ) );
							writer.Write( "ok" );
						}
						else if( msg.StartsWith( "retract" ) ) {
							DoRetract( StripWithDot( msg ) );
							writer.Write( "ok" );
						}
						else if( msg.Equals( "connect." ) ) {
							Console.WriteLine( "Connect." );
							HandleConnect( reader, writer,
								socketStream, connection );
						}
						else if( msg.StartsWith( "disconnect" ) ) {
							int toDisconnect = Int32.Parse( Strip( msg ) );
							agents[ toDisconnect ] = null;
							writer.Write( "ok" );
						}
						else if( msg.StartsWith( "synch" ) ) {
							DoSynch( reader, writer, Strip( msg ) );
						}
					} catch( Exception ) {
					} finally {
						try { reader.Close( ); } catch { }
						try { writer.Close( ); } catch { }
						try { socketStream.Close( ); } catch { }
						try { connection.Close( ); } catch { }
						//try { listener.Stop( ); } catch { }
						Console.WriteLine( "Closed connection.\n" );
					}	
				}
			} catch( Exception e ) {
				Console.WriteLine( "Error in runServer:\n" + e );
			} finally {
				listener.Stop( );
			}
		}
	}

	class AgentConnection {

		int id;

		// queue of strings representing asserts/retracts that should be
		// sent to this agent, the next time it synchronizes.
		private Queue pendingChanges = new Queue( );

		public void Enqueue( string s ) {
			this.pendingChanges.Enqueue( s );
		}

		public void SendQueue( BinaryWriter tr ) {
			string next;
			while( this.pendingChanges.Count > 0 ) {
				next = (string)( this.pendingChanges.Dequeue( ) );
				tr.Write( next );
				Console.WriteLine( "Broadcast to {0}: {1}", id, next );
			}
			tr.Write( "!commands." );
		}

		public AgentConnection( int id ) {
			this.id = id;
		}
	}

}