/*
 * @(#)dollar_abs_2.cs
 * Prolog predicate $abs/2
 */

namespace JJC.Psharp.Predicates {

	using JJC.Psharp.Lang;

	public sealed class dollar_abs_2 : Predicate {
		Term arg1, arg2;

		public dollar_abs_2(Term a1, Term a2, Predicate cont){
			arg1 = a1;
			arg2 = a2;
			this.cont = cont;
		}
		public dollar_abs_2(){}
		public override void setArgument(Term[] args, Predicate cont){
			arg1 = args[0];
			arg2 = args[1];
			this.cont = cont;
		}

		public override Predicate exec( Prolog engine ){
			Term a1, a2, num;
			a2 = arg2.Dereference();
			
			try {
				a1 = engine.EvalTerm(arg1);
				if (a1.IsInteger()) {
					int x = ((IntegerTerm)a1).IntValue();
					num = new IntegerTerm(System.Math.Abs(x));
				} else if (a1.IsDouble()) {
					double x = ((DoubleTerm)a1).DoubleValue();
					num = new DoubleTerm(System.Math.Abs(x));
				} else {
					return engine.fail();
				}
			} catch (ArithException) {
				return engine.abort("{TYPE ERROR: " + 
					this + 
					" -arg 1: expected expression, found " +
					arg1 + "}\n");
			}
			if( !a2.Unify(num, engine.trail) ) 
				return engine.fail();
			return cont;
		}

		public override int arity() { return 2; }

		public override string ToString() {
			return "$abs(" + arg1 + "," + arg2 + ")";
		}
	}

}