/*
 * @(#) dollar_close_1.cs
 * Prolog predicate $close/1 (close/1)
 */

namespace JJC.Psharp.Predicates {

	using JJC.Psharp.Lang;

	using System.IO;

	public sealed class dollar_close_1 : Predicate {
		public static SymbolTerm 
			SYM_user = SymbolTerm.MakeSymbol("user");
		public static SymbolTerm 
			SYM_user_input  = SymbolTerm.MakeSymbol("user_input");
		public static SymbolTerm 
			SYM_user_output = SymbolTerm.MakeSymbol("user_output");
		public static SymbolTerm 
			SYM_user_error = SymbolTerm.MakeSymbol("user_error");
		Term arg1;

		public dollar_close_1(Term a1, Predicate cont) {
			arg1 = a1;
			this.cont = cont;
		}
		public dollar_close_1() {}
		public override void setArgument(Term[] args, Predicate cont){
			arg1 = args[0];
			this.cont = cont;
		}

		public override Predicate exec( Prolog engine ) {
			Term a1;
			a1 = arg1.Dereference();
			object obj;
	
			try {
				if ( a1.IsVariable() ) {
					return engine.abort("{INSTANTIATION ERROR: " + this + "}\n");
				}
				if ( a1.Equals(SYM_user) ) {
					//engine.user_input.Close();
					//engine.user_output.Close();
					//engine.user_error.Close();
					return cont;
				}
				if ( a1.Equals(SYM_user_input) ) {
					//engine.user_input.Close();
					return cont;
				}
				if ( a1.Equals(SYM_user_output) ) {
					//engine.user_output.Close();
					return cont;
				}
				if ( a1.Equals(SYM_user_error) ) {
					//engine.user_error.Close();
					return cont;
				}
			} catch (IOException) {
				return engine.abort("{IO ERROR: " + this + "}\n");
			}

			if ( !a1.IsCsObject() ) {
				return engine.abort("{TYPE ERROR: " + this + 
					" -arg 1: expected stream, found " +
					a1 + "}\n");
			}
			obj = ((CsObjectTerm)a1).Object();
			try {
				if ( obj is TextWriter ) {
					((TextWriter)obj).Close();
					return cont;
				} 
				if ( obj is PushbackReader ) {
					((PushbackReader)obj).Close();
					return cont;
				} 
			} catch (IOException) {
				return engine.abort("{IO ERROR: " + this + "}\n");
			}
			return engine.abort("{TYPE ERROR: " + this + 
				" -arg 1: expected stream, found " +
				a1 + "}\n");
		}

		public override int arity() { return 1; }

		public override string ToString() {
			return "$close(" + arg1 + ")";
		}
	}

}