/*
 * dollar_functor_3.cs
 * Prolog predicate $functor/3
 */

namespace JJC.Psharp.Predicates {

	using JJC.Psharp.Lang;

	public sealed class dollar_functor_3 : Predicate {
		public static SymbolTerm SYM_dot = SymbolTerm.MakeSymbol(".");
		public static SymbolTerm Nil = SymbolTerm.MakeSymbol("[]");
		Term arg1, arg2, arg3;

		public dollar_functor_3(Term a1, Term a2, Term a3, Predicate cont) {
			arg1 = a1; 
			arg2 = a2; 
			arg3 = a3;
			this.cont = cont;
		}
		public dollar_functor_3() {}
		public override void setArgument(Term[] args, Predicate cont){
			arg1 = args[0];
			arg2 = args[1];
			arg3 = args[2];
			this.cont = cont;
		}

		public override Predicate exec( Prolog engine ) {
			Term a1, a2, a3;
			a1 = arg1.Dereference();
			a2 = arg2.Dereference();
			a3 = arg3.Dereference();

			Term functor, arity;
			Term[] args;

			if (a1.IsVariable()) {
				if ( !a3.IsInteger() ) {
					return engine.abort("{TYPE ERROR: " + this +
						" -arg 3: expected integer, found " +
						a3 + "}\n");
					//		return null;
				}
				int n = ((IntegerTerm)a3).value();
				if (n < 0) {
					return engine.abort("{DOMAIN ERROR: " + this +
						" -arg 3}\n");
					//		return null;
				}
				if (n == 0) {
					if ( !a2.IsSymbol() && !a2.IsInteger() && !a2.IsDouble() ) {
						return engine.abort("{TYPE ERROR: " + this +
							" -arg 2: expected atomic, found " +
							a2 + "}\n");
						//		    return null;
					}
					if( !a1.Unify(a2, engine.trail) )
						return engine.fail();
					return cont;
				}
				if ( !a2.IsSymbol() ) {
					return engine.abort("{TYPE ERROR: " + this +
						" -arg 2: expected atom, found " +
						a2 + "}\n");
					//		return null;
				}
				if ( n == 2 && a2.Equals(SYM_dot) ) {
					//		Term t = new ListTerm(new VariableTerm(), new VariableTerm());
					Term t = new ListTerm(engine.makeVariable(), engine.makeVariable());
					if( !a1.Unify(t, engine.trail) )
						return engine.fail();
					return cont;
				}
				args = new Term[n];
				for(int i=0; i<n; i++)
					args[i] = engine.makeVariable();
				//		args[i] = new VariableTerm();
				SymbolTerm sym = SymbolTerm.MakeSymbol(((SymbolTerm)a2).name, n);
				if( !a1.Unify(new StructureTerm(sym, args), engine.trail) )
					return engine.fail();
				return cont;
			} else {
				if ( a1.IsSymbol() || a1.IsInteger() || a1.IsDouble() ){
					functor = a1;
					arity   = new IntegerTerm(0);
				} else if ( a1.IsList() ){
					functor = SYM_dot;
					arity   = new IntegerTerm(2);
				} else if ( a1.IsStructure() ){
					//		functor = ((StructureTerm)a1).functor;
					functor = SymbolTerm.MakeSymbol(((StructureTerm)a1).functor.name);
					arity   = new IntegerTerm(((StructureTerm)a1).arity);
				} else {
					return engine.fail();
				}
				if( !a2.Unify(functor, engine.trail) ) 
					return engine.fail();
				if( !a3.Unify(arity, engine.trail) ) 
					return engine.fail();
			}
			return cont;
		}

		public override int arity() { return 3; }

		public override string ToString() {
			return "$functor(" + arg1 + ", " + arg2 + ", " + arg3 + ")";
		}
	}

}