/*
 * @(#)dollar_min_3.cs
 * Prolog predicate $min/3
 */

namespace JJC.Psharp.Predicates {

	using JJC.Psharp.Lang;

	public sealed class dollar_min_3 : Predicate {
		Term arg1, arg2, arg3;

		public dollar_min_3(Term a1, Term a2, Term a3, Predicate cont){
			arg1 = a1;
			arg2 = a2;
			arg3 = a3;
			this.cont = cont;
		}
		public dollar_min_3(){}
		public override void setArgument(Term[] args, Predicate cont){
			arg1 = args[0];
			arg2 = args[1];
			arg3 = args[2];
			this.cont = cont;
		}

		public override Predicate exec(Prolog engine){
			Term a1,a2,a3;
			Term num;

			a3 = arg3.Dereference();
			try {
				a1 = engine.EvalTerm(arg1);
				a2 = engine.EvalTerm(arg2);
				if (a1.IsInteger()) {
					if (a2.IsInteger()) {
						int x = ((IntegerTerm)a1).IntValue();
						int y = ((IntegerTerm)a2).IntValue();
						num = new IntegerTerm(System.Math.Min(x,y));
					} else if (a2.IsDouble()) {
						double x = ((IntegerTerm)a1).DoubleValue();
						double y = ((DoubleTerm)a2).DoubleValue();
						num = new DoubleTerm(System.Math.Min(x,y));
					} else {
						return engine.fail();
					}
				} else if (a1.IsDouble()) {
					if (a2.IsNumber()) {
						double x = ((DoubleTerm)a1).DoubleValue();
						double y = ((NumberTerm)a2).DoubleValue();
						num = new DoubleTerm(System.Math.Min(x,y));
					} else {
						return engine.fail();
					}
				} else {
					return engine.fail();
				}
			} catch (ArithException) {
				return engine.abort("{TYPE ERROR: " + this +  "}\n");
			}
			if(!a3.Unify(num, engine.trail)) 
				return engine.fail();
			return cont;
		}

		public override int arity() { return 3; }
		public override string ToString() {
			return "$min(" + arg1 + "," + arg2 + "," + arg3 + ")";
		}
	}

}