/*
 * @(#)dollar_runOnGlobalEngine_1.cs
 * Prolog predicate $run_on_global_engine/1
 */

namespace JJC.Psharp.Predicates {

	using JJC.Psharp.Lang;
	using System.Collections;

	public sealed class GlobalEngineCallback : Predicate {
		public dollar_runOnGlobalEngine_1 p;

		public GlobalEngineCallback( dollar_runOnGlobalEngine_1 p ){
			this.p = p;
		}

		public override Predicate exec( Prolog engine ) {
			p.success();
			return engine.fail();
		}

		public override void setArgument( Term[] t, Predicate p ) {
			throw new System.Exception( );
		}

		public override string ToString(){ return "GlobalEngineCallback";}
		public override int arity() { return 0; }
	}

	public sealed class dollar_runOnGlobalEngine_1 : Predicate {
		public static SymbolTerm SYM_true = SymbolTerm.MakeSymbol("true");
		public static SymbolTerm SYM_fail = SymbolTerm.MakeSymbol("fail");
		public static PrologInterface globalEngineInterface = new PrologInterface( true, false );

		Term arg1;

		public dollar_runOnGlobalEngine_1(Term a1, Predicate cont) {
			arg1 = a1;
			this.cont = cont;
		}
		public dollar_runOnGlobalEngine_1() {}
		public override void setArgument(Term[] args, Predicate cont) {
			arg1 = args[0];
			this.cont = cont;
		}

		private ArrayList solutions = new ArrayList( );
		private Term[] clonedArgs;
		private Prolog engine;

		public void success( ) {
			// now add the current instantiations to the list of results.
			Term[] clonedBackArgs = new Term[ clonedArgs.Length ];
			globalEngineInterface.engine.CloneReset( );
			for( int i = 0; i < clonedArgs.Length; i++ ) {
				clonedBackArgs[ i ] = globalEngineInterface.engine.CloneTerm( clonedArgs[ i ], engine, false );
			}
			solutions.Add( clonedBackArgs );
		}

		public override Predicate exec( Prolog engine ) {
			string pred_name = null;
			SymbolTerm functor = null;
			Term[] args = null;
			int arity;

			Term t = arg1.Dereference();
			//Term a2 = arg2.Dereference();

			if ( t.IsSymbol() ){
				functor = (SymbolTerm) t;
				arity = 0;
				// took out stuff for true/fail (try fork(fail,..) sometime)
				pred_name = functor.name;
				args = new Term[ 0 ];
			}
			else if ( t.IsStructure() ){
				functor = ((StructureTerm) t).functor;
				args  = ((StructureTerm) t).args;
				arity = ((StructureTerm) t).arity;
				pred_name = functor.name;      
			}
			else if ( t.IsVariable() ){
				return engine.abort("{INSTANTIATION ERROR: " + this +
					" -arg 1}\n");
			}
			else {
				return engine.abort("{TYPE ERROR: " + this +
					" -arg 1: expected goal, found " +
					t + "}\n");
			}

			Predicate pred;
			string class_name;

			class_name = getClassName(pred_name, arity);

			System.Type clazz;

			// new plan: we grab all the instantiations by using a callback,
			// (similar to ReturnCs), at end we add to the stack of calls
			// a call to assert each instantiation on the current engine
			// of the form '$soln'( <functo>( <final argument tuple> ) ).  

			// This is followed by:
			// clause( '$soln'(X)

			lock( this.GetType( ) ) {
				try {
					solutions.Clear( );
					
					clazz = AssemblyList.GetAnyType( class_name );	
					pred = (Predicate)System.Activator.CreateInstance( clazz );

					Term[] clonedArgs = new Term[ args.Length ];
					engine.CloneReset( );
					for( int i = 0; i < args.Length; i++ )
						clonedArgs[ i ] = engine.CloneTerm( args[ i ], globalEngineInterface.engine, false );

					pred.setArgument( clonedArgs, new GlobalEngineCallback( this ) ); 
				
					globalEngineInterface.SetPredicate( pred );
					this.clonedArgs = clonedArgs;
					this.engine = engine;

					// run on _this_ thread, we just can't create a new thread
					// everytime something concurrent gets instantiated.
					globalEngineInterface.Run( ); 
				
					Predicate fullcont = cont;
					foreach( Term[] ta in solutions ) {
						fullcont = new Predicates.Asserta_1( 
							new StructureTerm( SymbolTerm.MakeSymbol( "$soln", 1 ),
							new Term[] { new StructureTerm( functor, ta ) } ), fullcont );								
					}

					return new Predicates.Abolish_1( 
						new StructureTerm( 
						SymbolTerm.MakeSymbol( "/", 2 ),
						new Term[] { 
									   SymbolTerm.MakeSymbol( "$soln", 1 ), 
									   new IntegerTerm( 1 )
						} ), fullcont );
				} catch( System.Exception e ) {
					engine.user_output.WriteLine( e.Message );
					return engine.fail();
				}
			}

		}

			public override int arity() { return 1; }

		public override string ToString() {
			return "$run_on_global_engine(" + arg1 + ")";
		}

		/* This has been changed to generate the new style names. 
		 * JJC */

		public string getClassName( string s, int arity ) {
			char[] ch;
			int len;
			string s1 = "";
			
			ch = s.ToCharArray( );
			len = ch.Length;

			if( System.Char.IsLower( ch[ 0 ] ) )
				s1 += System.Char.ToUpper( ch[ 0 ] );
			else if( ch[ 0 ] == '_' )
				s1 += "__";
			else if( System.Char.IsDigit( ch[ 0 ] ) || 
				System.Char.IsUpper( ch[ 0 ] ) )
				s1 += "_" + ch[ 0 ];
			else 
				s1 += Token.GetReplacement( (int)ch[ 0 ] );

			int i = 1;
			while( i < len ) {
				if( i == len - 1 && ch[ i ] == '_' ) {
					s1 += "_underscore";
					break;
				} else if( System.Char.IsLower( ch[ i ] ) ||
					System.Char.IsDigit( ch[ i ] ) )
					s1 += ch[ i ];
				else if( System.Char.IsUpper( ch[ i ] ) )
					s1 += "_" + ch[ i ];
				else if( ch[ i ] == '_' ) {
					i++;
					if( System.Char.IsUpper( ch[ i ] ) ||
						System.Char.IsDigit( ch[ i ] ) )
						s1 += "__" + ch[ i ];
					else if( System.Char.IsLower( ch[ i ] ) )
						s1 += System.Char.ToUpper( ch[ i ] );
					else if( ch[ i ] == '_' )
						s1 += "__underscore";
					else 
						s1 += "__" + Token.GetReplacement( (int)ch[ i ] );
				} else
					s1 += "_" + Token.GetReplacement( (int)ch[ i ] );
				i++;
			}

			return "JJC.Psharp.Predicates." + s1 + "_" + arity;
		}
	}
}