/*
 * @(#)dollar_waitOn_1.cs
 * Prolog predicate $wait_on/1
 */

namespace JJC.Psharp.Predicates {

	using JJC.Psharp.Lang;

	public sealed class dollar_waitOn_1 : Predicate {
		Term arg1;

		public dollar_waitOn_1(Term a1, Predicate cont) {
			arg1 = a1;
			this.cont = cont;
		}
		public dollar_waitOn_1() {}
		public override void setArgument(Term[] args, Predicate cont){
			arg1 = args[0];
			this.cont = cont;
		}

		public override Predicate exec( Prolog engine ) {
			Term a1 = arg1.Dereference( );
			if( !a1.IsCsObject( ) ) 
				return engine.fail( );
			CsObjectTerm cso = (CsObjectTerm)a1;

			ConcurrentVariable cv = (ConcurrentVariable)( cso.ToCsObject( ) );

			lock( cv ) {
				if( cv.ShouldRelease( engine ) ) {
					return engine.fail( );
				}
				cv.DeclareWaiting( engine, true );
				if( cv.AllWaiting( ) ) {
					cv.DeclareWaiting( engine, false );
					return engine.fail( );
				} else {
					System.Threading.Monitor.Wait( cv );
					cv.DeclareWaiting( engine, false );
					if( cv.ShouldRelease( engine ) ) {
						return engine.fail( );
					}	
				}
			}
			return cont;
		}

		public override int arity() { return 1; }
		public override string ToString() {
			return "$wait_on(" + arg1 + ")";
		}	
	}

}