namespace JJC.Psharp.Lang {

	using System;
	using System.Threading;
	using System.Collections;

	public sealed class ConcurrentVariable {
		
		private ArrayList activeThreads =
			ArrayList.Synchronized( new ArrayList( ) );

		internal CsObjectTerm csObject = null;

		internal ConcurrentVariable( ) {
			csObject = new CsObjectTerm( this );
		}

		internal void Pulse( ) {
			Monitor.PulseAll( this );
		}

		internal bool AllWaiting( ) {
			foreach( ActiveThread at in activeThreads ) 
				if( !at.currentlyWaiting )
					return false;
			// now: all are waiting, so we mark them as pending release
			foreach( ActiveThread at in activeThreads )
				at.awaitingWaitRelease = true;
			Pulse( );
			return true;
		}

		internal void DeclareWaiting( Prolog engine, bool waiting ) {
			ActiveThread found = null;
			foreach( ActiveThread at in activeThreads )
				if( at.threadDescriptor == engine )
					found = at;
			if( found != null ) {
				found.currentlyWaiting = waiting;
				lock( this )
					Pulse( );
			}
		}

		internal bool ShouldRelease( Prolog engine ) {
			ActiveThread found = null;
			foreach( ActiveThread at in activeThreads )
				if( at.threadDescriptor == engine )
					found = at;
			if( found != null ) {
				if( found.awaitingWaitRelease ) {
					found.awaitingWaitRelease = false;
					lock( this )
						Pulse( );
					return true;
				} else
					return false;
			}
			return false;
		}

		internal void ReportThreadDeath( Prolog died ) {
			ActiveThread found = null;
			foreach( ActiveThread at in activeThreads )
				if( at.threadDescriptor == died )
					found = at;
			if( found != null ) {
				activeThreads.Remove( found );
				lock( this )
					Pulse( );
			}
		}

		internal void ReportFork( Prolog born ) {
			activeThreads.Add( new ActiveThread( born ) );
		}

		public override string ToString( ) {
			return "RootSynch: " + this.GetHashCode( );
		}
	}

	internal sealed class ActiveThread {
		private bool awaitingWaitReleaseInternal = false;
		private bool currentlyWaitingInternal = false;
	
		internal bool awaitingWaitRelease {
			set { 
				lock( this )
					awaitingWaitReleaseInternal = value;
			}
			get {
				lock( this )
					return awaitingWaitReleaseInternal;
			}
		}

		internal bool currentlyWaiting {
			set { 
				lock( this )
					currentlyWaitingInternal = value;
			}
			get {
				lock( this )
					return currentlyWaitingInternal;
			}
		}

		internal Prolog threadDescriptor = null;

		internal ActiveThread( Prolog threadDescriptor ) {
			this.threadDescriptor = threadDescriptor;
		}			
	}

	internal sealed class NewBinding {
		internal CsObjectTerm vt;
		internal Term b;
		internal NewBinding( CsObjectTerm vt, Term b ) {
			this.vt = vt;
			this.b = b;
		}
	}

}