/*
 *  @(#)CpfEntry.cs
 *  Enrty of Choice Point Frame
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 * Modified by Jon Cook.
 */

namespace JJC.Psharp.Lang {

	using System;
	using System.Collections;

	sealed internal class CpfEntry {
		public long timeStamp;
		public Term[] args;
		public Predicate cont;  // continuation 
		public Predicate bp;    // next cluase
		public int tr;          // trail pointer
		public int b0;          // cut point
		public int R, L, U;
		public Term R1, R2;

		// for speed up...
		private CpfEntry next = null;

		public static CpfEntry makeCpfEntry(Term[] args, Predicate cont,
					CpfEntry head ){
			if (head == null){
				return new CpfEntry(args, cont);
			}
			CpfEntry entry = head;
			head = head.next;
			entry.args = args;
			entry.cont = cont;
			return entry;
		}

		public void RemoveCpfEntry( CpfEntry head ){
			this.args = null;
			this.next = head;
			head = this;
		}

		// */
		
		private CpfEntry(Term[] args, Predicate cont){
			this.args = args;
			this.cont = cont;
		}
	}

	/*
	   *  @(#)CpfStack.cs
	 *  Choice Point Stack
	 */

	/* 
	 * @author Mutsunori Banbara 
	 *         banbara@pascal.seg.kobe-u.ac.jp
	 *         Nara National College of Technology
	 * @author Naoyuki Tamura 
	 *         tamura@kobe-u.ac.jp
	 *         Kobe University
     * Modified by Jon Cook.
	 */


	internal sealed class CpfStack {
		private int maxContents = 20000;
		private Stack buffer;
		private CpfEntry topEntry = null;
		private int itop;
		public int top {
			get {
				return itop;
			}
			set {
				itop = value;
			}
		}

		private CpfEntry head = null;

		public CpfStack( ){
			buffer = new Stack( maxContents ); 
			top = -1;
			head = null;
		}
		public CpfStack( int n ){
			maxContents = n;
			buffer = new Stack( maxContents ); 
			top = -1;
			head = null;
		}

		public void create(Term[] args, Predicate p){
			top++;
			buffer.Push( topEntry = CpfEntry.makeCpfEntry(args, p, head) );
		}

		public void Delete( ) { 
			top--;
			(((CpfEntry)(buffer.Pop( )))).RemoveCpfEntry( head );
			topEntry = (CpfEntry)buffer.Peek( );
		}

		public void Cut(int i){
			while (top > i) {
				top--;
				((CpfEntry)(buffer.Pop( ))).RemoveCpfEntry( head );
			}
			topEntry = (CpfEntry)buffer.Peek( );
		}

		public void SetTimeStamp(long n){ topEntry.timeStamp = n; }
		public long GetTimeStamp(){ return topEntry.timeStamp; }

		public void SetBP(Predicate p){ topEntry.bp = p; }
		public void SetTR(int i){ topEntry.tr = i; }
		public void SetB0(int i){ topEntry.b0 = i; }
		public Term[] GetArgs(){ return topEntry.args; }
		public Predicate GetCont(){ return topEntry.cont; }
		public Predicate GetBP(){ return topEntry.bp; }
		public int GetTR(){ return topEntry.tr; }
		public int GetB0(){ return topEntry.b0; }

		public Term GetR1(){ return topEntry.R1; }
		public Term GetR2(){ return topEntry.R2; }
		public void SetR1(Term i){ topEntry.R1 = i; }
		public void SetR2(Term i){ topEntry.R2 = i; }

		public int GetR(){ return topEntry.R; }
		public void SetR(int i){ topEntry.R = i; }
		public int GetL(){ return topEntry.L; }
		public void SetL(int i){ topEntry.L = i; }
		public int GetU(){ return topEntry.U; }
		public void SetU(int i){ topEntry.U = i; }

		public bool Empty(){ return top == -1; }
		public int Max(){ return maxContents; }
	}
}