/*
 *  @(#)CsObjectTerm.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 * Modified by Jon Cook.
 */
// import java.lang.*;

namespace JJC.Psharp.Lang {

	using System;
	using System.Collections;

	[System.Serializable]
	public sealed class CsObjectTerm : Term {
		public static readonly SymbolTerm Nil = SymbolTerm.MakeSymbol("[]");
		private object obj;
		private Type type;

		public CsObjectTerm(object o) { 
			obj = o; 
			type = obj.GetType( );
		}

		public void Undo() { obj = null; }

		public void Set(object o) { 
			obj = o; 
			type = obj.GetType( ); // obj.getClass(); TOOK OUT
		}

		public object Object() { return obj; }

		public override bool Unify(Term t, Trail trail) {
			t = t.Dereference();
			if ( t.IsVariable() )
				return ((VariableTerm)t).Unify(this, trail);
			if ( !t.IsCsObject() )
				return false;
			return obj.Equals(((CsObjectTerm)t).Object());
		}

		public override bool Equals(Term t) {
			//	t = t.Dereference();
			if( !t.IsCsObject() )
				return false;
			return obj.Equals(((CsObjectTerm)t).Object());
		}

		public override Term CopyTerm(Prolog engine) {
			return new CsObjectTerm(obj);
		}

		/* START JJC */
		internal override Term CloneTerm(Prolog oldEngine, Prolog newEngine, bool makeConc ) {
			return new CsObjectTerm( obj );
		}
		/* END JJC */

		public override object ToCsObject() {
			return obj;
		}

		public Term ToPrologTerm() {
			return CsTerm(obj);
		}

		private Term CsTerm(object o) {
			if( o is Int32 )
				return new IntegerTerm( (Int32)o );
			if( o is Double )
				return new DoubleTerm( (Double)o );
			if( o is string )
				return SymbolTerm.MakeSymbol( (string)o );
			if( o is Boolean )
				return SymbolTerm.MakeSymbol( o.ToString( ) );
			if( o is Array ) {
				object[] oo = (Object[])o;
				Term t = Nil;
				foreach( object ooo in oo )
					t = new ListTerm( (Term)ooo, t );
				return t;
			}
			return new CsObjectTerm( o );
		}

		public Type GetCsType() {
			return type;
		}

		public override string ToString() {
			if (obj == null)
				return "CsObjectTerm(null)";
			return type.FullName + "(" + obj.GetHashCode( ) + ")";
		}
	}

}