/*
 *  @(#)DoubleTerm.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 * Modified by Jon Cook.
 */

namespace JJC.Psharp.Lang {

	[System.Serializable]
	public sealed class DoubleTerm : NumberTerm {
		private double val;

		public DoubleTerm(double i) {	val = i; }

		public double value() { return val; }

		public override bool Unify( Term t, Trail trail ) {
			t = t.Dereference();
			if ( t.IsVariable() )
				return ((VariableTerm)t).Unify(this, trail);
			if ( !t.IsDouble() )
				return false;
			return val == ((DoubleTerm)t).value();
		}

		public override bool Equals(Term t) {
			if( !t.IsDouble() )
				return false;
			return val == ((DoubleTerm)t).value();
		}

		public override Term CopyTerm(Prolog engine) {
			return new DoubleTerm(val);
		}

		/* START JJC */
		internal override Term CloneTerm(Prolog oldEngine, Prolog newEngine, bool makeConc ) {
			return new DoubleTerm(val);
		}
		/* END JJC */

		public override object ToCsObject() {
			return val;
		}

		public override string ToString() {
			return val.ToString( );
		}

		public override int IntValue( ) {
			return (int)( System.Math.Floor( val ) );
		}

		public override long LongValue( ) {
			return (long)val;
		}

		public override double DoubleValue( ) {
			return val;
		}
	}

}