/*
 *  @(#)ListTerm.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 * Modified by Jon Cook.
 */

namespace JJC.Psharp.Lang {

	using System;

	[System.Serializable]
	public sealed class ListTerm : Term {
		private Term icar;
		public Term car {
			get {
				return icar;
			}
			set {
				icar = value;
			}
		}
		private Term icdr;
		public Term cdr {
			get {
				return icdr;
			}
			set {
				icdr = value;
			}
		}
		public ListTerm(Term _car, Term _cdr) { 
			car = _car; 
			cdr = _cdr; 
		}

		public override bool Unify(Term t, Trail trail) {
			t = t.Dereference();
			if ( t.IsVariable() )
				return ((VariableTerm)t).Unify(this, trail);
			if ( !t.IsList() )
				return false;
			ListTerm t1 = (ListTerm)t;
			return car.Unify(t1.car, trail) && cdr.Unify(t1.cdr, trail);
		}

		public override bool Equals(Term t) {
			//	t = t.Dereference();
			if ( !t.IsList() )
				return false;
			ListTerm t1 = (ListTerm)t;
			return car.Equals(t1.car.Dereference()) && cdr.Equals(t1.cdr.Dereference());
		}

		public override Term CopyTerm(Prolog engine) {
			return new ListTerm(car.CopyTerm(engine), cdr.CopyTerm(engine));
		}

		/* START JJC */
		internal override Term CloneTerm(Prolog oldEngine, Prolog newEngine, bool makeConc ) {
			return new ListTerm(car.CloneTerm( oldEngine, newEngine, makeConc ), 
				cdr.CloneTerm( oldEngine, newEngine, makeConc ));
		}
		/* END JJC */

		// $B$"$^$j$h$/$J$$!*(B
		public int Length() {
			int count = 0;
			Term tmp = this;
			while( tmp.IsList() ){
				count++;
				tmp = ((ListTerm)tmp).cdr.Dereference();
			}
			return count;
		}

		[ObsoleteAttribute("length is obsolete, use Length instead")]
		public int length( ) { return Length( ); }

		public override object ToCsObject() {
			int len = Length();
			object[] array = null;
			try {
				array = new Object[ len ]; //Array.newInstance(typeof( object ), len); // put this back 25/9/02
				Term tmp = this;
				object element = null;
				for (int i=0; i<len; i++) {
					element = ((ListTerm)tmp).car.Dereference().ToCsObject();
					array[ i ] = element; // Array.set(array, i, element); // put this back 25/9/02
					tmp = ((ListTerm)tmp).cdr.Dereference();
				}
			} catch (System.Exception e){
				System.Console.WriteLine( e.Message );
				System.Console.WriteLine("{ERROR : " + this + "}");
			}
			return (object)array;
		}

		public override string ToString() {
			string s;
			Term x = this;

			s = "[";
			for (;;) {
				s += ((ListTerm)x).car.ToString();
				x = ((ListTerm)x).cdr.Dereference();
				if (!(x is ListTerm))
					break;
				s += ",";
			}
			if (! x.IsNil()) {
				s += "|" + x.ToString();
			}
			s += "]";
			return s;
		}
	}

}