/*
 * @(#) Prolog.cs
 * Prolog engine & runtime environment
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 * Modified by Jon Cook.
 */

namespace JJC.Psharp.Lang {

	using JJC.Psharp.Lang.Resource;

	using System;
	using System.Collections;
	using System.IO;
	using System.Text;

	public class Prolog {
    
		internal readonly long runtime0 = System.DateTime.Now.Ticks / 10000; //System.currentTimeMillis();
		internal long old_runtime = 0;
    
		public static readonly SymbolTerm Nil = SymbolTerm.MakeSymbol("[]");

		/* START JJC */

		private Hashtable lockRecord = new Hashtable( );

		internal void ReportLock( object o ) {
			if( lockRecord[ o ] == null )
				lockRecord[ o ] = 1;
			else 
				lockRecord[ o ] = (int)lockRecord[ o ] + 1;
		}

		internal void ReportUnlock( object o ) {
			if( (int)( lockRecord[ o ] ) == 1 )
				lockRecord[ o ] = null;
			else
				lockRecord[ o ] = (int)lockRecord[ o ] - 1;
		}

		internal void ReleaseAllLocks( ) {
			IDictionaryEnumerator ide = null;
			try {
				ide = lockRecord.GetEnumerator( );
				while( ide.MoveNext( ) ) {
					try {
						for( int i = 0; i < (int)( ide.Value ); i++ )
							System.Threading.Monitor.Exit( ide.Key );
					} catch { }
				}
			} catch {
			}
		}

		internal Queue concInstQueue = null; // new Queue( );  OPT: null if not used.
		internal ArrayList concurrentVars = new ArrayList( );

		public readonly PushbackReader user_input;
		public readonly System.IO.TextWriter user_output;
		public readonly System.IO.TextWriter user_error;
		
		/* END JJC */

		public PushbackReader current_input; //= user_input;
		public TextWriter current_output; //= user_output;
    
		public readonly Term[] aregs = new Term[255];    
		public Predicate cont = null;       
  
		/* JJC: moved this from gen code to here, so that 
		 * multithreading is possible.   Should be ok given
		 * nature of first call optimisation */
		// public FcoPredicate entry_code = null; 

		/* public static IComparer comparer = new SubclassCompare( ); */
		public Hashtable entryCodes = new Hashtable( ); // null, comparer );

		public void SetEntryCode( Type t, FcoPredicate f ) {
			entryCodes[ t ] = f;
		}

		public FcoPredicate GetEntryCode( Type t ) {
			return (FcoPredicate)( entryCodes[ t ] );
		}

		internal readonly CpfStack stack = new CpfStack(); // Choice point frame
		public readonly Trail trail = new Trail();       // Trail
		internal int B0; // = stack.top;                  // Cut

		public readonly Hashtable user_table = new Hashtable();
		public Hashtable current_table = new Hashtable();

		public long timeStamp = System.Int64.MinValue;

		public readonly ResourceTable res = new ResourceTable();
		public int L =  1;         // $B8=:_>CHq2DG=$J%j%=!<%9$N%l%Y%k(B
		public int U = -1;         // $B>CHq8e$K%;%C%H$5$l$kCM(B
		public int R0; // = res.top; // $B4XO"%j%9%H:n@.$K;HMQ(B(begin_imp,more_imp)
		public Term R1;            // $B>CHq2DG=$J(Bindex$B$NA4%j%9%H(B(by hash)
		public Term R2;            // $B>CHq2DG=$J(Bindex$B$N%j%9%H(B($B%j%=!<%9DI2C;~$KBh(B1$B0z?t$,JQ?t(B)

		public Prolog( PushbackReader new_in, System.IO.TextWriter new_out,
			System.IO.TextWriter new_err ) {
			R0 = res.top;
			B0 = stack.top;

			user_input = new_in;
			user_output = new_out;
			user_error = new_err;

			current_input = user_input;
			current_output = user_output;
			res.setEngine(this);
			trail.SetEngine(this);
			init();
		}

		public Prolog() { 
			R0 = res.top;
			B0 = stack.top;

			user_input = new PushbackReader( System.Console.In );
			user_output	= System.Console.Out;
			user_error = System.Console.Error;

			current_input = user_input;
			current_output = user_output;
			res.setEngine(this);
			trail.SetEngine(this);
			init();
		}

		public void init() { 
			Term[] noarg = {};
			stack.create(noarg, null);
			stack.SetTR(trail.top);
			stack.SetTimeStamp(++timeStamp);
			stack.SetBP(null);    
			stack.SetB0(B0);
			stack.SetR(res.top);
			stack.SetL(L);
			stack.SetU(U);
		}

		public VariableTerm makeVariable(){
			return new VariableTerm(timeStamp, this); // "this" added by JJC
		}

		// $B3F<o%G!<%?%(%j%"$N>pJs<hF@(B
		public int stack_top(){ return stack.top; }
		public int stack_max(){ return stack.Max(); }
		public int trail_top(){ return trail.top; }
		public int trail_max(){ return trail.Max(); }
		public int resTop()   { return res.top; } 

		// $B9`$N%3%T!<(B(copy_term/2$B$G;HMQ(B)
		public Hashtable varTable = new Hashtable();
		public Term CopyTerm(Term t){
			t = t.Dereference();
			varTable.Clear();
			return t.CopyTerm(this);
		}

		/* START JJC */

		internal Hashtable cloneVarTable = new Hashtable( );

		internal void CloneReset( ) {
			cloneVarTable.Clear( );
		}

		internal Term CloneTerm( Term t, Prolog newEngine, bool makeConc ) {
		    t = t.Dereference( );
			return t.CloneTerm( this, newEngine, makeConc );
		}

		/* END JJC */

		// $B%F!<%V%k!"F~=PNO$N(Bcurrent$BCM$N@_Dj(B
		public void setTable(Hashtable table){
			current_table = table;
		}
		public void setInput(PushbackReader cin){
			current_input = cin;
		}
		public void setOutput(TextWriter cout){
			current_output = cout;
		}

		public Predicate abort(string s){ throw new AbortException(s); }

		public Predicate fail() {
			if ( stack.Empty() ) 
				return null;
			B0 = stack.GetB0();     // B0$B$r85$KLa$9(B
			return stack.GetBP();   // next clause$B$r<B9T(B
		}

		public void setB0()    { B0 = stack.top; }  // cut point$B$N%;%C%H(B
		public void neckCut()  { stack.Cut(B0); }     // neck cut$B$N<B9T(B
		public void cut(int i) { stack.Cut(i); }      // $BIaDL$N(Bcut$B$N<B9T(B

		public void restore(){
			Term[] args = stack.GetArgs();
			int i = args.Length;
			System.Array.Copy(args, 0, aregs, 1, i);
			cont = stack.GetCont();
		}

		// switch_on_term
		public Predicate switch_on_term(Predicate var, 
			Predicate num, 
			Predicate con, 
			Predicate str, 
			Predicate list) {
			Term arg1 = aregs[1].Dereference();
			if ( arg1.IsVariable() )  return var;
			if ( arg1.IsNumber() )    return num;
			if ( arg1.IsSymbol() )    return con;
			if ( arg1.IsStructure() ) return str;
			if ( arg1.IsList() )      return list;
			return fail();
		}

		// try
		public Predicate jtry(Predicate p, Predicate next){
			int i = p.arity();
			Term[] args = new Term[i];
			System.Array.Copy(aregs, 1, args, 0, i);
			stack.create(args, cont);
			stack.SetTR(trail.top);
			stack.SetTimeStamp(++timeStamp);
			stack.SetBP(next);
			stack.SetB0(B0);
			// for LLP
			stack.SetR(res.top);
			stack.SetL(L);
			stack.SetU(U);
			return p;
		}

		// retry
		public Predicate retry(Predicate p, Predicate next){
			restore();
			trail.Unwind(stack.GetTR());
			stack.SetBP(next);
			// for LLP
			res.unwind(stack.GetR());
			L = stack.GetL();
			U = stack.GetU();
			return p;
		}

		// trust
		public Predicate trust(Predicate p){
			restore();
			trail.Unwind(stack.GetTR());
			// for LLP
			res.unwind(stack.GetR());
			L = stack.GetL();
			U = stack.GetU();
			stack.Delete();
			return p;
		}


		//----------------------------------------------------------------- 
		//                        $B%j%=!<%9=hM}4XO"(B 
		//----------------------------------------------------------------- 

		// $B4XO"%j%=!<%9$NHO0O$N@8@.$K;HMQ(B
		public void setR0(){ R0 = res.top; }

		// $B3F<o%j%=!<%9%l%Y%k$X$NA`:n(B
		public void decrementU(){ U--; }
		public void incrementU(){ U++; }
		public void decrementL(){ L--; }
		public void incrementL(){ L++; }


		public void lookUpHash(SymbolTerm functor){
			res.setR1andR2(functor);
		}

		public bool pickupResource(SymbolTerm sym, int N){
			Term t;
			for (; R1.IsList(); R1 = ((ListTerm)R1).cdr){
				t = ((ListTerm)R1).car;
				if ( !t.IsResIndex() ){
					current_output.WriteLine("error in pickupResource\n");
					return false;
				}
				if ( res.consumable(sym, ((ResIndexTerm)t).getIndex()) ){
					aregs[N] = t;
					R1 = ((ListTerm)R1).cdr;
					return true;
				}
			}
			for (; R2.IsList(); R2 = ((ListTerm)R2).cdr){
				t = ((ListTerm)R2).car;
				if ( !t.IsResIndex() ){
					current_output.WriteLine("error in pickupResource\n");
					return false;
				}
				if ( res.consumable(sym, ((ResIndexTerm)t).getIndex()) ){
					aregs[N] = t;
					R2 = ((ListTerm)R2).cdr;
					return true;
				}
			}
			return false;
		}

		// Linear$B%j%=!<%9$NDI2C(B
		public void addRes(Term head, ClosureTerm closure){
			res.addRes(head, closure);
		}
		// Exponential$B%j%=!<%9$NDI2C(B
		public void addExpRes(Term head, ClosureTerm closure){
			res.addExpRes(head, closure);
		}
		// R1, R2$B$N%;%C%H(B
		public void setRelatedResField(){
			res.setRelatedResField();
		}
		public void midWith(){ 
			res.midWith(); 
		}
		public bool endWith(){ 
			return res.endWith(); 
		}
		public bool endImp(ResIndexTerm i, ResIndexTerm j){
			return res.endImp(i,j);
		}
		public void endExpImp(ResIndexTerm i, ResIndexTerm j){
			res.endExpImp(i, j);
		}

		// try resource
		public Predicate tryResource(Predicate p, Predicate next){
			int i;  /*  num_of_args */
			Term[] args;

			i = p.arity();
			args = new Term[i];
			System.Array.Copy(aregs, 1, args, 0, i);
			stack.create(args, cont);
			stack.SetTR(trail.top);
			stack.SetTimeStamp(++timeStamp); //added
			stack.SetBP(next);
			stack.SetB0(B0);
			stack.SetR1(R1);
			stack.SetR2(R2);
			stack.SetR(res.top);
			stack.SetL(L);
			stack.SetU(U);
			return p;
		}

		// choice point frame$B$+$i%l%8%9%?$NCM$rLa$9(B
		public void restoreResource(){
			restore();
			trail.Unwind(stack.GetTR());
			R1 = stack.GetR1();
			R2 = stack.GetR2();
			res.unwind(stack.GetR());
			L = stack.GetL();
			U = stack.GetU();
		}

		// R1,R2$B$r8=:_$NCM$K99?7$7!"(BBP$B$b%;%C%H$7D>$9(B
		public Predicate retryResource(Predicate p, Predicate next){
			stack.SetR1(R1);
			stack.SetR2(R2);
			stack.SetBP(next);
			return p;
		}

		// choice point frame$B$r:o=|(B
		public Predicate trustResource(Predicate p){
			stack.Delete();
			return p;
		}

		// $B%j%=!<%9$K>CHq%^!<%/$rIU$1$k(B
		public ClosureTerm consume(int N){
			if ( ! aregs[N].IsResIndex() )
				current_output.WriteLine("error in consume/1 in Prolog.cs");
			int index = ((ResIndexTerm)(aregs[N])).getIndex();
			res.consume(index);
			return res.getClosureAt(index);
		}

		// closure$B$N<B9T(B
		public Predicate executeClosure(ClosureTerm closure){
			Predicate code = closure.code;
			Term[] vars = closure.freeVars;
			int len = closure.length;
			if ( len == 0 ) 
				return code;
			System.Array.Copy(vars, 0, aregs, code.arity()+1, len);
			return code;
		}

		// $B%*%W%F%#%^%$%:MQ(B
		public bool hasMoreResource(){
			Term t;
			for (; R1.IsList(); R1 = ((ListTerm)R1).cdr){
				t = ((ListTerm)R1).car;
				if ( !t.IsResIndex() ){
					current_output.WriteLine("error in noResource\n");
					return true;
				}
				if ( res.consumable(((ResIndexTerm)t).getIndex()) ){
					break;
				}
			}
			for (; R2.IsList(); R2 = ((ListTerm)R2).cdr){
				t = ((ListTerm)R2).car;
				if ( !t.IsResIndex() ){
					current_output.WriteLine("error in noResource\n");
					return true;
				}
				if ( res.consumable(((ResIndexTerm)t).getIndex()) ) {
					break;
				}
			}
			if ( R1.IsNil() ){
				R1 = R2;
				R2 = Nil;
			}
			return R1.IsList();
		}

		// top$B%4!<%k$N<B9T(B
		public void executeTop(){ 
			res.executeTop(); 
		}

    /** A constant holding the string representation of "(+)/1". */
    protected static SymbolTerm SYM_PLUS_1      = SymbolTerm.MakeSymbol("+", 1);
    /** A constant holding the string representation of "(+)/2". */
    protected static SymbolTerm SYM_PLUS_2      = SymbolTerm.MakeSymbol("+", 2);
    /** A constant holding the string representation of "(-)/1". */
    protected static SymbolTerm SYM_MINUS_1     = SymbolTerm.MakeSymbol("-", 1);
    /** A constant holding the string representation of "(-)/2". */
    protected static SymbolTerm SYM_MINUS_2     = SymbolTerm.MakeSymbol("-", 2);
    /** A constant holding the string representation of "(*)/2". */
    protected static SymbolTerm SYM_TIMES_2     = SymbolTerm.MakeSymbol("*", 2);
    /** A constant holding the string representation of "(/)/2". */
    protected static SymbolTerm SYM_FLOATQUOT_2 = SymbolTerm.MakeSymbol("/", 2);
    /** A constant holding the string representation of "(//)/2". */
    protected static SymbolTerm SYM_INTQUOT_2   = SymbolTerm.MakeSymbol("//", 2);
    /** A constant holding the string representation of "(mod)/2". */
    protected static SymbolTerm SYM_MOD_2       = SymbolTerm.MakeSymbol("mod", 2);
    /** A constant holding the string representation of "(max)/2". */
    protected static SymbolTerm SYM_MAX_2       = SymbolTerm.MakeSymbol("max", 2);
    /** A constant holding the string representation of "(min)/2". */
    protected static SymbolTerm SYM_MIN_2       = SymbolTerm.MakeSymbol("min", 2);
    /** A constant holding the string representation of "(>>)/2". */
    protected static SymbolTerm SYM_RSHIFT_2    = SymbolTerm.MakeSymbol(">>", 2);
    /** A constant holding the string representation of "(<<)/2". */
    protected static SymbolTerm SYM_LSHIFT_2    = SymbolTerm.MakeSymbol("<<", 2);
    /** A constant holding the string representation of "(/\)/2". */
    protected static SymbolTerm SYM_BITCONJ_2   = SymbolTerm.MakeSymbol("/\\", 2);
    /** A constant holding the string representation of "(\/)/2". */
    protected static SymbolTerm SYM_BITDISJ_2   = SymbolTerm.MakeSymbol("\\/", 2);
    /** A constant holding the string representation of "(\)/1". */
    protected static SymbolTerm SYM_BITNEG_1    = SymbolTerm.MakeSymbol("\\", 1);
    /** A constant holding the string representation of "(#)/2". */
    protected static SymbolTerm SYM_BITXOR_2    = SymbolTerm.MakeSymbol("#", 2);
    /** A constant holding the string representation of "(abs)/1". */
    protected static SymbolTerm SYM_ABS_1       = SymbolTerm.MakeSymbol("abs", 1);
    /** A constant holding the string representation of "(integer)/1". */
    protected static SymbolTerm SYM_INT_1       = SymbolTerm.MakeSymbol("integer", 1);
    /** A constant holding the string representation of "(float)/1". */
    protected static SymbolTerm SYM_FLOAT_1     = SymbolTerm.MakeSymbol("float", 1);
    /** A constant holding the string representation of "random". */
    protected static SymbolTerm SYM_RANDOM      = SymbolTerm.MakeSymbol("random");
    /** A constant holding the string representation of "pi". */
    protected static SymbolTerm SYM_PI          = SymbolTerm.MakeSymbol("pi");
    /** A constant holding the string representation of "e". */
    protected static SymbolTerm SYM_E           = SymbolTerm.MakeSymbol("e");
    /** A constant holding the string representation of "(sin)/1". */
    protected static SymbolTerm SYM_SIN_1       = SymbolTerm.MakeSymbol("sin", 1);
    /** A constant holding the string representation of "(cos)/1". */
    protected static SymbolTerm SYM_COS_1       = SymbolTerm.MakeSymbol("cos", 1);
    /** A constant holding the string representation of "(tan)/1". */
    protected static SymbolTerm SYM_TAN_1       = SymbolTerm.MakeSymbol("tan", 1);
    /** A constant holding the string representation of "(asin)/1". */
    protected static SymbolTerm SYM_ASIN_1      = SymbolTerm.MakeSymbol("asin", 1);
    /** A constant holding the string representation of "(acos)/1". */
    protected static SymbolTerm SYM_ACOS_1      = SymbolTerm.MakeSymbol("acos", 1);
    /** A constant holding the string representation of "(atan)/1". */
    protected static SymbolTerm SYM_ATAN_1      = SymbolTerm.MakeSymbol("atan", 1);
    /** A constant holding the string representation of "(sqrt)/1". */
    protected static SymbolTerm SYM_SQRT_1      = SymbolTerm.MakeSymbol("sqrt", 1);
    /** A constant holding the string representation of "(log)/1". */
    protected static SymbolTerm SYM_LOG_1       = SymbolTerm.MakeSymbol("log", 1);
    /** A constant holding the string representation of "(exp)/1". */
    protected static SymbolTerm SYM_EXP_1       = SymbolTerm.MakeSymbol("exp", 1);
    /** A constant holding the string representation of "(pow)/2". */
    protected static SymbolTerm SYM_POW_2       = SymbolTerm.MakeSymbol("pow", 2);
    /** A constant holding the string representation of "(degrees)/1". */
    protected static SymbolTerm SYM_DEGREES_1 = SymbolTerm.MakeSymbol("degrees", 1);
    /** A constant holding the string representation of "(radians)/1". */
    protected static SymbolTerm SYM_RADIANS_1 = SymbolTerm.MakeSymbol("radians", 1);

	private System.Random random = new System.Random( );

    protected bool isFloat = false;

		public Term EvalTerm(Term t) {
			t = t.Dereference();
			if (t.IsNumber())
				return t;
			else if (t == SYM_RANDOM)
				return new DoubleTerm(random.NextDouble());
			else if (t == SYM_PI)
				return new DoubleTerm(System.Math.PI);
			else if (t == SYM_E)
			
				return new DoubleTerm(System.Math.E);
			else if (t.IsList())
				return EvalTerm(((ListTerm)t).car);
			else if (!t.IsStructure())
				throw (new PrologException()); 
			// arithmethic expression
			isFloat = false;
			double val = DoubleValue(t);
			if (isFloat)
				return new DoubleTerm(val);
			return new IntegerTerm((int)val);
		}

		public double DoubleValue(Term t) {
			t = t.Dereference();
			if (t.IsInteger()) {
				return ((IntegerTerm)t).DoubleValue();
			} else if (t.IsDouble()) {
				isFloat = true;
				return ((DoubleTerm)t).DoubleValue();
			} else if (t == SYM_RANDOM) {
				isFloat = true;
				return random.NextDouble();
			} else if (t == SYM_PI) {
				isFloat = true;
				return Math.PI;
			} else if (t == SYM_E) {
				isFloat = true;
				return Math.E;
			} else if (t.IsList())
				return DoubleValue(((ListTerm)t).car);
			else if (!t.IsStructure()) {
				throw (new PrologException()); 
			}
			SymbolTerm func;
			Term[] args;
			int arity;
			func  = ((StructureTerm)t).functor;
			args  = ((StructureTerm)t).args;
			arity = ((StructureTerm)t).arity;

			if (func == SYM_PLUS_1) {
				return DoubleValue(args[0]);
			} else if (func == SYM_PLUS_2) {
				bool flag0 = isFloat;
				isFloat = false;
				double d1 = DoubleValue(args[0]);
				bool flag1 = isFloat;
				isFloat = false;
				double d2 = DoubleValue(args[1]);
				isFloat = (flag0 || flag1 || isFloat);
				return (d1 + d2);
			} else if (func == SYM_MINUS_1) {
				return - DoubleValue(args[0]);
			} else if (func == SYM_MINUS_2) {
				bool flag0 = isFloat;
				isFloat = false;
				double d1 = DoubleValue(args[0]);
				bool flag1 = isFloat;
				isFloat = false;
				double d2 = DoubleValue(args[1]);
				isFloat = (flag0 || flag1 || isFloat);
				return (d1 - d2);
			} else if (func == SYM_TIMES_2) {
				bool flag0 = isFloat;
				isFloat = false;
				double d1 = DoubleValue(args[0]);
				bool flag1 = isFloat;
				isFloat = false;
				double d2 = DoubleValue(args[1]);
				isFloat = (flag0 || flag1 || isFloat);
				return (d1 * d2);
			} else if (func == SYM_INTQUOT_2) {
				bool flag0 = isFloat;
				int d1 = (int)(DoubleValue(args[0]));
				int d2 = (int)(DoubleValue(args[1]));
				isFloat = (flag0 || false);
				return (double)(d1 / d2);
			} else if (func == SYM_FLOATQUOT_2) {
				bool flag0 = isFloat;
				double d1 = DoubleValue(args[0]);
				double d2 = DoubleValue(args[1]);
				isFloat = (flag0 || true);
				return (d1 / d2);
			} else if (func == SYM_MOD_2) {
				bool flag0 = isFloat;
				double d1 = DoubleValue(args[0]);
				double d2 = DoubleValue(args[1]);
				isFloat = (flag0 || false);
				return (double)( (int)(d1 % d2) );
			} else if (func == SYM_ABS_1) {
				return System.Math.Abs(DoubleValue(args[0]));
			} else if (func == SYM_MAX_2) {
				bool flag0 = isFloat;
				isFloat = false;
				double d1 = DoubleValue(args[0]);
				bool flag1 = isFloat;
				isFloat = false;
				double d2 = DoubleValue(args[1]);
				isFloat = (flag0 || flag1 || isFloat);
				return System.Math.Max(d1, d2);
			} else if (func == SYM_MIN_2) {
				bool flag0 = isFloat;
				isFloat = false;
				double d1 = DoubleValue(args[0]);
				bool flag1 = isFloat;
				isFloat = false;
				double d2 = DoubleValue(args[1]);
				isFloat = (flag0 || flag1 || isFloat);
				return System.Math.Min(d1, d2);
			} else if (func == SYM_RSHIFT_2) {
				bool flag0 = isFloat;
				double d1 = DoubleValue(args[0]);
				double d2 = DoubleValue(args[1]);
				isFloat = (flag0 || false);
				return (double)((int)d1 >> (int)d2);
			} else if (func == SYM_LSHIFT_2) {
				bool flag0 = isFloat;
				double d1 = DoubleValue(args[0]);
				double d2 = DoubleValue(args[1]);
				isFloat = (flag0 || false);
				return (double)((int)d1 << (int)d2);
			} else if (func == SYM_BITNEG_1) {
				bool flag0 = isFloat;
				double d = DoubleValue(args[0]);
				isFloat = (flag0 || false);
				return  (double)(~ ((int)d));
			} else if (func == SYM_BITCONJ_2) {
				bool flag0 = isFloat;
				double d1 = DoubleValue(args[0]);
				double d2 = DoubleValue(args[1]);
				isFloat = (flag0 || false);
				return (double)((int)d1 & (int)d2);
			} else if (func == SYM_BITDISJ_2) {
				bool flag0 = isFloat;
				double d1 = DoubleValue(args[0]);
				double d2 = DoubleValue(args[1]);
				isFloat = (flag0 || false);
				return (double)((int)d1 | (int)d2);
			} else if (func == SYM_BITXOR_2) {
				bool flag0 = isFloat;
				double d1 = DoubleValue(args[0]);
				double d2 = DoubleValue(args[1]);
				isFloat = (flag0 || false);
				return (double)((int)d1 ^ (int)d2);
			} else if (func == SYM_INT_1) {
				bool flag0 = isFloat;
				double d = DoubleValue(args[0]);
				isFloat = (flag0 || false);
				return (double)( Math.Floor(d) );
			} else if (func == SYM_FLOAT_1) {
				double d = DoubleValue(args[0]);
				isFloat = true;
				return d;
			} else if (func == SYM_SIN_1) {
				bool flag0 = isFloat;
				double d = DoubleValue(args[0]);
				isFloat = (flag0 || true);
				return System.Math.Sin(d);
			} else if (func == SYM_COS_1) {
				bool flag0 = isFloat;
				double d = DoubleValue(args[0]);
				isFloat = (flag0 || true);
				return System.Math.Cos(d);
			} else if (func == SYM_TAN_1) {
				bool flag0 = isFloat;
				double d = DoubleValue(args[0]);
				isFloat = (flag0 || true);
				return System.Math.Tan(d);
			} else if (func == SYM_ASIN_1) {
				bool flag0 = isFloat;
				double d = DoubleValue(args[0]);
				isFloat = (flag0 || true);
				return System.Math.Asin(d);
			} else if (func == SYM_ACOS_1) {
				bool flag0 = isFloat;
				double d = DoubleValue(args[0]);
				isFloat = (flag0 || true);
				return System.Math.Acos(d);
			} else if (func == SYM_ATAN_1) {
				bool flag0 = isFloat;
				double d = DoubleValue(args[0]);
				isFloat = (flag0 || true);
				return System.Math.Atan(d);
			} else if (func == SYM_SQRT_1) {
				bool flag0 = isFloat;
				double d = DoubleValue(args[0]);
				isFloat = (flag0 || true);
				return System.Math.Sqrt(d);
			} else if (func == SYM_LOG_1) {
				bool flag0 = isFloat;
				double d = DoubleValue(args[0]);
				isFloat = (flag0 || true);
				return System.Math.Log(d);
			} else if (func == SYM_EXP_1) {
				bool flag0 = isFloat;
				double d = DoubleValue(args[0]);
				isFloat = (flag0 || true);
				return System.Math.Exp(d);
			} else if (func == SYM_POW_2) {
				bool flag0 = isFloat;
				double d1 = DoubleValue(args[0]);
				double d2 = DoubleValue(args[1]);
				isFloat = (flag0 || true);
				return System.Math.Pow(d1,d2);
			} else if (func == SYM_DEGREES_1) {
				bool flag0 = isFloat;
				double d = DoubleValue(args[0]);
				isFloat = (flag0 || true);
				return d * 180.0 / Math.PI;
			} else if (func == SYM_RADIANS_1) {
				bool flag0 = isFloat;
				double d = DoubleValue(args[0]);
				isFloat = (flag0 || true);
				return d / 180.0 * Math.PI;
			} else {
				throw (new ArithException());
			}
		}
		//---------------------------------------------------------------------
		//                           $B9`$NHf3S(B
		//---------------------------------------------------------------------
		
	    protected static SymbolTerm SYM_DOT_2 = SymbolTerm.MakeSymbol(".", 2);
		public static int EQUAL  = 0;
		public static int BEFORE = -1;
		public static int AFTER  = 1;

		// compare function
		// Variables < Integers < Atoms < Compound terms(containing lists)
		
		public int CompareTerm(Term _a1, Term _a2) {
			Term a1, a2;
			int rc;            // return code
			SymbolTerm f1, f2; // functor
			Term[] ag1, ag2;   // arguments
			int ar1, ar2;      // arity
			Term t1, t2;
	
			a1 = _a1.Dereference();
			a2 = _a2.Dereference();
			if (a1 == a2) 
				return EQUAL;
			if (a1.IsVariable()) {
				if (! a2.IsVariable())
					return BEFORE;
				long i = ((VariableTerm)a1).GetID();
				long j = ((VariableTerm)a2).GetID();
				if (i < j) 
					return BEFORE;
				return AFTER;
			}
			if (a1.IsDouble()) {
				if (a2.IsVariable()) 
					return AFTER;
				if (! a2.IsDouble())
					return BEFORE;
				double i = ((DoubleTerm)a1).DoubleValue();
				double j = ((DoubleTerm)a2).DoubleValue();
				return (int)(i - j);
			}
			if (a1.IsInteger()) {
				if (a2.IsVariable() || a2.IsDouble())
					return AFTER;
				if (! a2.IsInteger()) 
					return BEFORE;
				int i = ((IntegerTerm)a1).IntValue();
				int j = ((IntegerTerm)a2).IntValue();
				return (i - j);
			}
			if (a1.IsSymbol()) {
				if (a2.IsVariable() || a2.IsNumber())
					return AFTER;
				if (! a2.IsSymbol()) 
					return BEFORE;
				String i = ((SymbolTerm)a1).name;
				String j = ((SymbolTerm)a2).name;
				return i.CompareTo(j) ;
			}
			if (a1.IsList()) {
				if (a2.IsVariable() || a2.IsNumber() || a2.IsSymbol())
					return AFTER;
				if (a2.IsStructure()) {
					f2  = ((StructureTerm)a2).functor;
					ag2 = ((StructureTerm)a2).args;
					ar2 = ((StructureTerm)a2).arity;
					if (2 != ar2)
						return (2 - ar2);
					if (SYM_DOT_2 != f2)
						return CompareTerm(SYM_DOT_2, f2);
					t1 = ((ListTerm)a1).car;
					t2 = ag2[0];
					rc = CompareTerm(t1, t2);
					if (rc != EQUAL)
						return rc;
					t1 = ((ListTerm)a1).cdr;
					t2 = ag2[1];
					return CompareTerm(t1, t2);
				}
				if (!a2.IsList())
					return BEFORE;
				t1 = ((ListTerm)a1).car;
				t2 = ((ListTerm)a2).car;
				for (int i=0; i<2; i++) {
					rc = CompareTerm(t1, t2);
					if (rc != EQUAL)
						return rc;
					t1 = ((ListTerm)a1).cdr;
					t2 = ((ListTerm)a2).cdr;
				}
				return EQUAL;
			}
			if (a1.IsStructure()) {
				if (a2.IsVariable() || a2.IsNumber() || a2.IsSymbol() )
					return AFTER;
				f1  = ((StructureTerm)a1).functor;
				ar1 = ((StructureTerm)a1).arity;
				ag1 = ((StructureTerm)a1).args;
				if (a2.IsList()) {
					if (ar1 != 2) 
						return (ar1 - 2);
					if (f1 != SYM_DOT_2)
						return CompareTerm(f1, SYM_DOT_2);
					t1 = ag1[0];
					t2 = ((ListTerm)a2).car;
					rc = CompareTerm(t1, t2);
					if (rc != EQUAL)
						return rc;
					t1 = ag1[1];
					t2 = ((ListTerm)a2).cdr;
					return CompareTerm(t1, t2);
				}
				if (!a2.IsStructure())
					return BEFORE;
				f2  = ((StructureTerm)a2).functor;
				ar2 = ((StructureTerm)a2).arity;
				ag2 = ((StructureTerm)a2).args;
				if (ar1 != ar2 )
					return (ar1 - ar2);
				if (f1 != f2)
					return CompareTerm(f1, f2);
				for (int i=0; i<ar1; i++) {
					rc = CompareTerm(ag1[i], ag2[i]);
					if (rc != EQUAL) 
						return rc;
				}
				return EQUAL;
			}
			if (a1.IsCsObject()) {
				if (!a2.IsCsObject())
					return AFTER;
				return (a1.ToString()).CompareTo(a2.ToString());
			}
			return AFTER;
		}

		/* START JJC */
		
		internal Predicate ServiceQueue( Predicate cont ) {
			if( this.concInstQueue != null ) {
				while( this.concInstQueue.Count != 0 ) {
					NewBinding b = (NewBinding)( concInstQueue.Dequeue( ) );
					Predicate fullCont = new Predicates.dollar_unlock_1( b.vt, cont );
					fullCont = new Predicates.dollar_pulseOn_1( b.vt, fullCont );
					fullCont = new Predicates.GlobalAssertz_1(
						new StructureTerm(
						SymbolTerm.MakeSymbol( "$msg", 2 ),
						new Term[] { b.vt, b.b } ), fullCont );
					fullCont = new Predicates.dollar_lock_1( b.vt, fullCont );
					return fullCont;
				} 
			}
			return cont;
		}
		/* END JJC */	
	}
}