/*
 * @(#)PrologMain.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 * Modified by Jon Cook.
 */

namespace JJC.Psharp.Lang {

	using System;
	using System.Reflection;
	
	using JJC.Psharp.Predicates;

	public sealed class PrologMain {

		/* entry point for Mono Loader */

		public static void CallbackMain( String[] argv, Assembly ass ) {
			AssemblyList.Add( ass, null );
			Main( argv );
		}

		/* main entry point */

		public static void Main( String[] argv ) {

			if( !Assembly.GetCallingAssembly( ).Equals( Assembly.GetExecutingAssembly( ) ) )
				AssemblyList.Add( Assembly.GetCallingAssembly( ), null );

			System.Console.WriteLine( "\nP# 1.1.3" );
			System.Console.WriteLine( "Copyright(C) 2003 J.J.Cook" );
			System.Console.WriteLine( "\nBased on:" );
			System.Console.WriteLine( "\n" + "Prolog Cafe 0.4.4, 0.6.1");
			System.Console.WriteLine("Copyright(C) 1997-2003 M.Banbara and N.Tamura");

			Prolog engine = new Prolog();
			Predicate cont = new Fail_0(null); 
			Predicate code;
			if (argv.Length == 0)
				code = makePredicate("Interpreter"); //$B%$%s%?%W%j%?(B
			else 
				code = makePredicate(argv[0]);
			code.setArgument(null, cont);
			try {
				while (true) {
					if (code == null) 
						break;
					code = code.exec( engine );
					code = engine.ServiceQueue( code );
				}
			}
			catch (AbortException e){
				System.Console.WriteLine(e.Message);
			}		
		}

		public static Predicate makePredicate(String pred){
			String className = "JJC.Psharp.Predicates." + pred + "_" + 0;
			Type clazz;
			Predicate code = null;
			try {
				clazz = AssemblyList.GetAnyType( className );	
				code = (Predicate)Activator.CreateInstance( clazz );
			} catch( Exception ) {
				System.Console.WriteLine("{EXISTENCE ERROR: procedure " + pred + "/0" +
					" does not exist}\n");
				throw new PrologException( );
			}
			return code;
		}

		private static readonly SymbolTerm Nil = SymbolTerm.MakeSymbol("[]");

		public static Term makeList(int n){
			Term t = Nil;
			for (int i=n; i>0; i--){
				t = new ListTerm(new IntegerTerm(i), t);
			}
			return t;
		}
	}
}