/*
 * @(#)PrologMainGUI.cs
 */

namespace JJC.Psharp.Lang {

	using System;
	using System.Reflection;
	using System.IO;
	
	using JJC.Psharp.Predicates;

	public sealed class PrologMainGUI {
	
		public static void Stop( ) {
			try {
				System.Threading.Thread.CurrentThread.Abort( );
			} catch( System.Threading.ThreadAbortException ) {
			}
		}

		public static void CallbackStart( PushbackReader uin, TextWriter uout, 
			TextWriter uerr, Assembly ass ) {
			AssemblyList.Add( ass, null );
			Start( uin, uout, uerr );
		}

		public static void Start( PushbackReader uin, TextWriter uout, TextWriter uerr ) {

			if( !Assembly.GetCallingAssembly( ).Equals( Assembly.GetExecutingAssembly( ) ) )
				AssemblyList.Add( Assembly.GetCallingAssembly( ), null );

			System.Console.WriteLine( "\nP# 1.1.3" );
			System.Console.WriteLine( "Copyright(C) 2002 J.J.Cook" );
			System.Console.WriteLine( "\nBased on:" );
			System.Console.WriteLine( "\n" + "Prolog Cafe 0.4.4, 0.6.1");
			System.Console.WriteLine("Copyright(C) 1997-2003 M.Banbara and N.Tamura");

			Prolog engine = new Prolog( uin, uout, uerr );
			Predicate cont = new Fail_0(null); 
			Predicate code;
			
			uout.WriteLine( "P# 1.1.3" );

			code = makePredicate("Interpreter"); //$B%$%s%?%W%j%?(B

			code.setArgument(null, cont);
			try {
				while (true) {
					if (code == null) 
						break;
					code = code.exec( engine );
					code = engine.ServiceQueue( code );
				}
			}
			catch (AbortException e){
				engine.user_output.WriteLine(e.Message);
			} catch( System.Threading.ThreadAbortException ) {
			}
		}

		public static Predicate makePredicate(String pred){
			String className = "JJC.Psharp.Predicates." + pred + "_" + 0;
			Type clazz;
			Predicate code = null;
			try {
				clazz = AssemblyList.GetAnyType( className );	
				code = (Predicate)Activator.CreateInstance( clazz );
			} catch( Exception ) {
				System.Console.WriteLine("{EXISTENCE ERROR: procedure " + pred + "/0" +
					" does not exist}\n");
				throw new PrologException( );
			}
			return code;
		}

		private static readonly SymbolTerm Nil = SymbolTerm.MakeSymbol("[]");

		public static Term makeList(int n){
			Term t = Nil;
			for (int i=n; i>0; i--){
				t = new ListTerm(new IntegerTerm(i), t);
			}
			return t;
		}
	}

}