/*
 * @(#)StructureTerm.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 * Modified by Jon Cook.
 */

namespace JJC.Psharp.Lang {

	[System.Serializable]
	public sealed class StructureTerm : Term {
		private SymbolTerm ifunctor;
		public SymbolTerm functor {
			get {
				return ifunctor;
			}
			set {
				ifunctor = value;
			}
		}

		private Term[] iargs;
		public Term[] args {
			get {
				return iargs;
			}
			set {
				iargs = value;
			}
		}

		private int iarity;
		public int arity {
			get {
				return iarity;
			}
			set {
				iarity = value;
			}
		}

		public StructureTerm(SymbolTerm _functor, Term[] _args){
			functor = _functor;
			args = _args;
			arity = functor.arity;
		}

		public string Name(){
			return functor.name;
		}

		public override Term CopyTerm(Prolog engine) {
			Term[] a = new Term[arity];
			for (int i = 0; i < arity; i++)
				a[i] = args[i].CopyTerm(engine);
			return new StructureTerm(functor, a);
		}

		/* START JJC */
		internal override Term CloneTerm( Prolog oldEngine, Prolog newEngine, bool makeConc ) {
			Term[] a = new Term[arity];
			for (int i = 0; i < arity; i++)
				a[i] = args[i].CloneTerm( oldEngine, newEngine, makeConc );
			return new StructureTerm(functor, a);
		}
		/* END JJC */

		public override bool Unify(Term t, Trail trail) {
			t = t.Dereference();
			if ( t.IsVariable() )
				return ((VariableTerm)t).Unify(this, trail);
			if ( !t.IsStructure() )
				return false;
			StructureTerm s = (StructureTerm)t;
			if ( functor != s.functor )
				return false;
			for (int i=0; i<arity; i++)
				if ( !args[i].Unify(s.args[i], trail) )
					return false;
			return true;
		}

		public override bool Equals(Term t) {
			Term a1, a2;
			if ( !t.IsStructure() )
				return false;
			StructureTerm s = (StructureTerm)t;
			if ( functor != s.functor )
				return false;
			for ( int i = 0; i < arity; i++ ){
				a1 = args[i].Dereference();
				a2 = s.args[i].Dereference();
				if ( ! a1.Equals(a2) )
					return false;
			}
			return true;
		}

		public override string ToString() {
			string s;
			string delim = "";
			s = functor.ToString() + "(";
			for (int i = 0; i < arity; i++) {
				s += delim + args[i].ToString();
				delim = ",";
			}
			s += ")";
			return s;
		}
	}
}