/*
 * @(#)SymbolTerm.cs 
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 * Modified by Jon Cook.
 */

namespace JJC.Psharp.Lang {

	using System;
	using System.Collections;

	[System.Serializable]
	public sealed class SymbolTerm : Term {

		private static Hashtable symbolTable = 
			Hashtable.Synchronized( new Hashtable( ) );

		private string iname;

	
		public string name {
			get {
				return iname;
			}
			set {
				iname = value;
			}
		}

		private int iarity;

		public int arity {
			get {
				return iarity;
			}
			set {
				iarity = value;
			}
		}

		// f/n$B7?$NA4%j%=!<%9$N(BIndex$B%j%9%H(B
		public Term res = Prolog.Nil;   
		// f/n$B7?$GBh0l0z?t$,JQ?t$N$b$N$N(BIndex$B%j%9%H(B
		public Term res2 = Prolog.Nil;  

		public static void ClearRes(){
			SymbolTerm sym = null;
			foreach( object symb in symbolTable ) {
				sym = (SymbolTerm)symb;
				sym.res  = Prolog.Nil;   
				sym.res2 = Prolog.Nil;
			}
		}

		[Obsolete( "Use MakeSymbol instead" )]
		public static SymbolTerm makeSymbol( string n ) { return MakeSymbol( n ); }
		[Obsolete( "Use MakeSymbol instead" )]
		public static SymbolTerm makeSymbol( string n, int a ) {
			return MakeSymbol( n, a );
		}

		public static SymbolTerm MakeSymbol(string _name){
			return MakeSymbol(_name, 0);
		}

		public static SymbolTerm MakeSymbol(string _name, int _arity){
			string key = _name + "/" + _arity;
			SymbolTerm sym = (SymbolTerm)symbolTable[key];
			if ( sym == null ){
				sym = new SymbolTerm(_name, _arity);
				symbolTable[ key ] = sym;
			}
			return sym;
		}

		private SymbolTerm(string _name, int _arity) { 
			name  = _name; 
			arity = _arity;
		}

		public override bool Unify(Term t, Trail trail) {
			t = t.Dereference();
			if ( t.IsVariable() )
				return ((VariableTerm)t).Unify(this, trail);
			if ( !t.IsSymbol() )
				return false;
			return name.Equals(((SymbolTerm)t).name);
		}

		public override bool Equals(Term t) {
			if ( !t.IsSymbol() )
				return false;
			return name.Equals(((SymbolTerm)t).name);
		}

		public void UndoRes(){
			if ( !res.IsList() )
				System.Console.WriteLine("SYSTEM ERROR: in undoRes");
			res = ((ListTerm)res).cdr;
		}

		public void UndoRes2(){
			if ( !res2.IsList() )
				System.Console.WriteLine("SYSTEM ERROR: in undoRes2");
			res2 = ((ListTerm)res2).cdr;
		}

		public override object ToCsObject() {
			return name;
		}

		public override string ToString() { return name; }
	}
}