/*
 * @(#)Term.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 * Modified by Jon Cook.
 * 
 */

namespace JJC.Psharp.Lang {

	using System;
	using JJC.Psharp.Lang.Resource;

	[System.Serializable]
	public abstract class Term {

		[Obsolete( "use Dereference instead" )]
		public Term dereference() { return Dereference(); }
		[Obsolete( "use IsNil instead" )]
		public bool isNil() { return IsNil(); }
		[Obsolete( "use IsVariable instead" )]
		public bool isVariable() { return IsVariable(); }
		[Obsolete( "use IsSymbol instead" )]
		public bool isSymbol() { return IsSymbol(); }
		[Obsolete( "use IsInteger instead" )]
		public bool isInteger() { return IsInteger(); }
		[Obsolete( "use IsList instead" )]
		public bool isList() { return IsList(); }
		[Obsolete( "use IsStructure instead" )]
		public bool isStructure() { return IsStructure(); }
		[Obsolete( "use IsCsObject instead" )]
		public bool isCsObject() { return IsCsObject(); }
		[Obsolete( "use IsResIndex instead" )]
		public bool isResIndex() { return IsResIndex(); }
		[Obsolete( "use IsClosure instead" )]
		public bool isClosure() { return IsClosure(); }
		[Obsolete( "use ToCsObject instead" )]
		public object toCsObject() { return ToCsObject(); }
		[Obsolete( "use CopyTerm instead" )]
		public Term copyTerm( Prolog engine ) { return CopyTerm( engine ); }
		[Obsolete( "use Unify instead" )]
		public bool unify( Term t, Trail trail ) { return Unify( t, trail ); }

		public virtual Term Dereference(){
			return this;
		}
	
		public bool IsNil(){
			return Prolog.Nil.Equals( this );
		}

		public bool IsVariable(){
			return this is VariableTerm;
		}

		public bool IsSymbol(){
			return this is SymbolTerm;
		}

		public bool IsNumber() {
			return this is NumberTerm;
		}

		public bool IsInteger(){
			return this is IntegerTerm;
		}
		
		public bool IsDouble() {
			return this is DoubleTerm;
		}

		public bool IsList(){
			return this is ListTerm;
		}

		public bool IsStructure(){
			return this is StructureTerm;
		}

		public bool IsCsObject(){
			return this is CsObjectTerm;
		}

		public bool IsResIndex(){
			return this is ResIndexTerm;
		}

		public bool IsClosure(){
			return this is ClosureTerm;
		}

		public virtual object ToCsObject(){
			return this;
		}

		public virtual Term CopyTerm(Prolog engine){ 
			return this; 
		}

		/* START JJC */
		internal virtual Term CloneTerm(Prolog oldEngine, Prolog newEngine, bool makeConc ) {
			return this; // careful!
		}
		/* END JJC */

		abstract public bool Unify( Term t, Trail trail );
		abstract public bool Equals( Term t );
	}
}