/*
 * @(#)Trail Stack
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 * Modified by Jon Cook.
 */

namespace JJC.Psharp.Lang {

	using System;
	using System.Collections;

	using JJC.Psharp.Lang.Resource;

	public sealed class Trail {
		private int maxContents = 20000;
		private Stack buffer;

		private int itop;
		public int top {
			get {
				return itop;
			}

			set {
				itop = value;
			}
		}
		internal Prolog engine;
	
		internal Trail() {
			buffer = new Stack( maxContents );
			top = -1;
		}
		internal Trail(int n){
			maxContents = n;
			buffer = new Stack( maxContents );
			top = -1;
		}

		internal void SetEngine(Prolog _engine){
			engine = _engine;
		}

		internal void Push(Term t){
			++top;
			buffer.Push( t );	
		}

		internal Term Pop(){
			Term t = (Term)buffer.Pop( );
			top--; 
			return t;
		}

		internal bool Empty(){
			return top == -1;
		}

		internal int Max(){
			return maxContents;
		}

		internal void Unwind(int i){
			Term t, s;
			String key;
			int n;
			int k,m;
	
			while ( top > i ){
				t = Pop();
				if ( t.IsVariable() ){
					((VariableTerm)t).Undo();
					continue;
				}
				if ( t.IsInteger() ){
					n = ((IntegerTerm)t).value();
					k = ((ResIndexTerm)Pop()).getIndex();
					engine.res.setLevel(k,n);
					continue;
				}
				if ( t.IsSymbol() ) {
					((SymbolTerm)t).UndoRes();
					s = Pop();
					if (! s.IsSymbol() ) Console.WriteLine("are????");
					if ( s == t )
						((SymbolTerm)s).UndoRes2();
					else{
						key = ((SymbolTerm)s).name;
						t = (Term)engine.res.hash[key];
						if (! t.IsList() ) Console.WriteLine("are????");
						engine.res.hash[ key ] = ((ListTerm)t).cdr.Dereference();
					}
					continue;
				}
				if ( t.IsResIndex() ){
					m = ((ResIndexTerm)t).getIndex();
					k = ((ResIndexTerm)Pop()).getIndex();
					for (; k<m; k++){
						engine.res.setOutOfScope(k, false);
					}
					continue;
				}
				Console.WriteLine("error in unwind trail");
			}
		}

		public void Show(){
			if ( Empty() ) {
				Console.WriteLine("Trail is empty!");
				return;
			}
			Console.WriteLine( this.ToString( ) );
		}
	}

}