/*
 *  @(#)ClosureTerm.cs
 */

/* 
 * @author Mutsunori Banbara 
 *         banbara@pascal.seg.kobe-u.ac.jp
 *         Nara National College of Technology
 * @author Naoyuki Tamura 
 *         tamura@kobe-u.ac.jp
 *         Kobe University
 * Modified by Jon Cook.
 */

namespace JJC.Psharp.Lang.Resource {

	using JJC.Psharp.Lang;

	public sealed class ClosureTerm : Term {

		private Predicate icode;

		public Predicate code {
			get {
				return icode;
			}
			set {
				icode = value;
			}
		}

		private Term[] ifreeVars;
		public Term[] freeVars {
			get {
				return ifreeVars;
			}
			set {
				ifreeVars = value;
			}
		}
		private int ilength;
		public int length {
			get {
				return ilength;
			}
			set {
				ilength = value;
			}
		}

		public ClosureTerm(Predicate _code, Term[] _freeVars) {
			code = _code;
			freeVars = _freeVars;
			length = freeVars.Length;
		}

		public override bool Unify(Term t, Trail trail) {
			t = t.Dereference();
			if ( t.IsVariable() )
				return ((VariableTerm)t).Unify(this, trail);
			if ( !t.IsClosure() )
				return false;
			ClosureTerm s = (ClosureTerm)t;
			if ( ! code.Equals(s.code) )
				return false;
			if (  length != s.length )
				return false;
			for (int i=0; i<length; i++)
				if ( !freeVars[i].Unify(s.freeVars[i], trail) )
					return false;
			return true;
		}

		public override bool Equals(Term t) {
			Term a1,a2;
			//	t = t.Dereference();
			if ( !t.IsClosure() )
				return false;
			ClosureTerm s = (ClosureTerm)t;
			if ( ! code.Equals(s.code) )
				return false;
			if (  length != s.length )
				return false;
			for ( int i=0; i<length; i++ ){
				a1 = freeVars[i].Dereference();
				a2 = s.freeVars[i].Dereference();
				if ( ! a1.Equals(a2) )
					return false;
			}
			return true;
		}

		public override Term CopyTerm(Prolog engine) {
			Term[] a = new Term[length];
			for (int i = 0; i < length; i++)
				a[i] = freeVars[i].CopyTerm(engine);
			return new ClosureTerm(code, a);
		}

		public override string ToString() {
			string s;
			string delim = "";
			s = "closure(" + code.ToString() + ", [";
			for (int i = 0; i < length; i++) {
				s += delim + freeVars[i].ToString();
				delim = ",";
			}
			s += "])";
			return s;
		}
	}

}

