/*
 * *** Please do not edit ! ***
 * @(#) Fit_3.cs
 * @procedure fit/3 in mltoprolog.pl
 */

/*
 * @version P# 1.1.3, on Sept 1 2003;  Prolog Cafe 0.44, on November 12 1999
 * @author Mutsunori Banbara (banbara@pascal.seg.kobe-u.ac.jp)
 * @author Naoyuki Tamura    (tamura@kobe-u.ac.jp)
 * Modified by Jonathan Cook (jjc@dcs.ed.ac.uk)
 */
namespace JJC.Psharp.Predicates {

using JJC.Psharp.Lang;
using JJC.Psharp.Lang.Resource;
using Predicates = JJC.Psharp.Predicates;
using Resources = JJC.Psharp.Resources;

public class Fit_3 : Predicate {
    static internal readonly Predicate Fit_3_1 = new Predicates.Fit_3_1();
    static internal readonly Predicate Fit_3_2 = new Predicates.Fit_3_2();
    static internal readonly Predicate Fit_3_var = new Predicates.Fit_3_var();

    public Term arg1, arg2, arg3;

    public Fit_3(Term a1, Term a2, Term a3, Predicate cont) {
        arg1 = a1; 
        arg2 = a2; 
        arg3 = a3; 
        this.cont = cont;
    }

    public Fit_3(){}
    public override void setArgument(Term[] args, Predicate cont) {
        arg1 = args[0]; 
        arg2 = args[1]; 
        arg3 = args[2]; 
        this.cont = cont;
    }

    public override Predicate exec( Prolog engine ) {
        engine.aregs[1] = arg1;
        engine.aregs[2] = arg2;
        engine.aregs[3] = arg3;
        engine.cont = cont;
        return call( engine );
    }

    public virtual Predicate call( Prolog engine ) {
        engine.setB0();
        return engine.switch_on_term(
                                   Fit_3_var,
                                   Fit_3_2,
                                   Fit_3_2,
                                   Fit_3_2,
                                   Fit_3_var
                                   );
    }

    public override int arity() { return 3; }

    public override string ToString() {
        return "fit(" + arg1 + ", " + arg2 + ", " + arg3 + ")";
    }
}

sealed class Fit_3_var : Fit_3 {
    static internal readonly Predicate Fit_3_var_1 = new Predicates.Fit_3_var_1();

    public override Predicate exec( Prolog engine ) {
        return engine.jtry(Fit_3_1, Fit_3_var_1);
    }
}

sealed class Fit_3_var_1 : Fit_3 {

    public override Predicate exec( Prolog engine ) {
        return engine.trust(Fit_3_2);
    }
}

sealed class Fit_3_1 : Fit_3 {
    static internal readonly SymbolTerm f1 = SymbolTerm.MakeSymbol("extent", 3);

    public override Predicate exec( Prolog engine ) {
        Term a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11;
        Predicate p1, p2, p3;
        a1 = engine.aregs[1].Dereference();
        a2 = engine.aregs[2].Dereference();
        a3 = engine.aregs[3].Dereference();
        Predicate cont = engine.cont;

        if ( a1.IsList() ){
            a4 = ((ListTerm)a1).car;
            a5 = ((ListTerm)a1).cdr;
        } else if ( a1.IsVariable() ){
            a4 = engine.makeVariable();
            a5 = engine.makeVariable();
            if ( !a1.Unify(new ListTerm(a4, a5), engine.trail) )
                return engine.fail();
        } else {
            return engine.fail();
        }
        if ( a4.IsStructure() ){
            if (f1 != ((StructureTerm)a4).functor )
                return engine.fail();
            Term[] args = ((StructureTerm)a4).args;
            a6 = args[1];
        } else if (a4.IsVariable() ){
            a6 = engine.makeVariable();
            Term[] args = {engine.makeVariable(), a6, engine.makeVariable()};
            if ( !a4.Unify(new StructureTerm(f1, args), engine.trail) )
                return engine.fail();
        } else {
            return engine.fail();
        }
        if ( a2.IsList() ){
            a7 = ((ListTerm)a2).car;
            a8 = ((ListTerm)a2).cdr;
        } else if ( a2.IsVariable() ){
            a7 = engine.makeVariable();
            a8 = engine.makeVariable();
            if ( !a2.Unify(new ListTerm(a7, a8), engine.trail) )
                return engine.fail();
        } else {
            return engine.fail();
        }
        if ( a7.IsStructure() ){
            if (f1 != ((StructureTerm)a7).functor )
                return engine.fail();
            Term[] args = ((StructureTerm)a7).args;
            a9 = args[0];
        } else if (a7.IsVariable() ){
            a9 = engine.makeVariable();
            Term[] args = {a9, engine.makeVariable(), engine.makeVariable()};
            if ( !a7.Unify(new StructureTerm(f1, args), engine.trail) )
                return engine.fail();
        } else {
            return engine.fail();
        }
        a10 = engine.makeVariable();
        a11 = engine.makeVariable();
        p1 = new Predicates.dollar_max_3(a10, a11, a3, cont);
        p2 = new Predicates.dollar_minus_3(a6, a9, a11, p1);
        p3 = new Predicates.Fit_3(a5, a8, a10, p2);
        return new Predicates.dollar_neckCut_0(p3);
    }
}

sealed class Fit_3_2 : Fit_3 {
    static DoubleTerm s1 = new DoubleTerm(0);

    public override Predicate exec( Prolog engine ) {
        Term a1, a2, a3;
        a1 = engine.aregs[1].Dereference();
        a2 = engine.aregs[2].Dereference();
        a3 = engine.aregs[3].Dereference();
        Predicate cont = engine.cont;

        if ( !s1.Unify(a3, engine.trail) ) return engine.fail();
        return cont;
    }
}
}

