:- dynamic( form1/1 ).

load_and_prime( Lock ) :-
	fork( load_and_prime_thread( Lock ) ).
	
load_and_prime_thread( X ) :-
	load_assembly( 'ClassHierarchy_database' ),
	lock( lock ),
	call( derives( C, D ) ),
%   write( C ), write( ' ' ), write( D ), nl,
	unlock( lock ), lock( lock ),
	fail.
load_and_prime_thread( X ) :-
	unlock( lock ).	
	
get_root( CentralClass, MaxUp, Root, PathToRoot, Hole ) :-
	get_root_inner( CentralClass, MaxUp, Root, 0, PathToRoot, Hole ).
	
get_root_inner( Class, MaxUp, Class, MaxUp, Z, Z ) :- !.
get_root_inner( Class, MaxUp, Root, Count, PathToRoot, PathSoFar ) :-
	call( derives( Superclass, Class ) ), !,
	Count1 is Count + 1,
	get_root_inner( Superclass, MaxUp, Root, Count1, PathToRoot, node( Superclass, [PathSoFar] ) ).
get_root_inner( Class, _, Class, _, Z, Z ) :- !.

generate_from_root( Root, Depth, Tree ) :-
	generate_from_root( Root, Depth, Tree, 0 ).

generate_from_root( Root, Depth, node( Root, [] ), Depth ) :- !.
generate_from_root( Root, Depth, node( Root, Children ), Count ) :-
	Count1 is Count + 1,
	findall( L , 
		generate_from_root_prime( Root, Depth, L, Count1 ),
		Children ).

generate_from_root_prime( Root, Depth, L, Count1 ) :-
	call( derives( Root, X ) ),
	validate( X ),
	generate_from_root( X, Depth, L, Count1 ).
	
generate_tree( CentralClass, Form1, Depth, Lock ) :-
	load_assembly( 'ClassHierarchy_database' ),
	assert( form1( Form1 ) ),
	lock( lock ),
	get_root( CentralClass, 10, Root, PathToRoot, RestOfTree ),
	% ( var( PathToRoot ) -> RealDepth = 0; RealDepth = Depth ),
	RealDepth = Depth,
	generate_from_root( CentralClass, RealDepth, RestOfTree ),
	Tree = PathToRoot,
	draw_tree( CentralClass, Tree ),
	unlock( lock ).

draw_tree( CentralClass, Tree ) :-
    design( Tree, extentf, pair( ResultTree, ResultExtent ) ),
    collapse_extent( ResultExtent, extent( Left, _, _ ) ), !,
    MinusLeft = -Left,
    draw2( MinusLeft, 0, ResultExtent, ResultTree, _ ),
    end_invokes.

extentf( S, extent( NHW, HW, 20 ) ) :-
    get_half_width( S, HW ),
    NHW is -HW.
    
/* ------------------------------------------------------------- 
   From file: Tree.sml
   ------------------------------------------------------------- */

% datatype 'a tree
% ----------------

% node( {'a}, [ {'a tree}, {'a tree}, {'a tree} ] ).

% fun movetree: node * real -> node
% ---------------------------------

movetree( node( pair( Label, X ), Subtrees ), 
	  Xprime, 
	  node( pair( Label, Xsum ), Subtrees ) ) :-
  Xsum is X + Xprime.

% type extent
% -----------

% extent( x, y, z ).

% fun moveextent
% --------------

moveextent_fn( extent( Left, Right, Height ),  %%CORRECT
	       X,
	       extent( NewLeft, NewRight, Height ) ) :-
  NewLeft is Left + X,
  NewRight is Right + X.

moveextent( [], _, [] ) :- !.
moveextent( [H1|T1], X, [H2|T2] ) :-
  moveextent_fn( H1, X, H2 ),
  moveextent( T1, X, T2 ).

% fun merge
% ---------

merge( [], QS, QS ).
merge( PS, [], PS ).
merge( [ extent( L1, _, H1 ) | PS ],
       [ extent( _, R1, H2 ) | QS ],
       [ extent( L1, R1, Max ) | Merge ] ) :-
  max( H1, H2, Max ),
  merge( PS, QS, Merge ).

% fun foldr/
% fun foldr f u [] = u | foldr f u (x::l) = f(x, foldr f u l);
% ------------------------------------------------------------

foldr( F, U, [], U ) :- !.  %% CORRECT
foldr( F, U, [X|L], FF ) :-
  foldr( F, U, L, Folded ),
  ToCall =.. [ F, X, Folded, FF ],
  call( ToCall ).

% fun mergelist
% -------------

mergelist( ES, Merged ) :-
  foldr( merge, [], ES, Merged ).

% fun fit
% -------

fit( [ extent( _, R, _ ) | PS ],
     [ extent( L, _, _ ) | QS ],
     Result ) :- !,
  fit( PS, QS, Fitted ),
  Diff is R - L,
  Result is max( Fitted, Diff ).
fit( _, _, 0.0 ).

% fun fitlistl_prime
% ------------------

fitlistl_prime( Acc, [], [] ) :- !.
fitlistl_prime( Acc, [E|ES], [X|Fitted] ) :-
  fit( Acc, E, X ),
  moveextent(E, X, Moved ),
  merge( Acc, Moved, Merged ),
  fitlistl_prime( Merged, ES, Fitted ).

% fun fitlistl
% ------------

fitlistl( ES, Result ) :-
  fitlistl_prime( [], ES, Result ).

% fun fitlistr_prime
% ------------------

fitlistr_prime( Acc, [], [] ) :- !.
fitlistr_prime( Acc, [E|ES], [X|Fitted] ) :-
  fit( E, Acc, Xneg ),
  X is -Xneg,
  moveextent( E, X, Moved ),
  merge( Moved, Acc, Merged ),
  fitlistr_prime( Merged, ES, Fitted ).	      

% fun fitlistr
% ------------

fitlistr( ES, Result ) :-
  reverse( ES, ESrev ),
  fitlistr_prime( [], ESrev, ResultReverse ),
  reverse( ResultReverse, Result ).

% fun mean
% --------

mean( X, Y, M ) :-
  M is ( X + Y ) / 2.

% fun fitlist
% -----------

fitlist( ES, Out ) :-
    fitlistl( ES, FittedLeft ),
    fitlistr( ES, FittedRight ),
    fitlist2( FittedLeft, FittedRight, Out ).
    
fitlist2( [], [], [] ) :- !.
fitlist2( [HL|TL], [HR|TR], [HM|TM] ) :-
	mean( HL, HR, HM ),
	fitlist2( TL, TR, TM ).
	
% fun Listpair.unzip
% ------------------
% [(a,b),(c,d),(e,f)] -> [a,c,e] * [b,d,f].

listpair_unzip( [], [], [] ) :- !.
listpair_unzip( [pair(P1,P2)|T1], [P1|T2], [P2|T3] ) :-
	listpair_unzip( T1, T2, T3 ).

% fun map/
% -------
% f * [a,b,c] -> [f(a), f(b), f(c)]

map( _, [], [] ) :- !.
map( F, [H1|T1], [H2|T2] ) :-
	ToCall =.. [ F, H1, H2 ],
	call( ToCall ),
	map( F, T1, T2 ).

% fun Listpair.map
% ----------------
% f * [a,b,c] * [d,e,f] -> [ f(a,d), f(b,e), f(c,f) ].

listpair_map( _, [], [], [] ) :- !.
listpair_map( F, [H1|T1], [H2|T2], [H3|T3] ) :-
    ToCall =.. [ F, H1, H2, H3 ],
	call( ToCall ),
	listpair_map( F, T1, T2, T3 ).

% fun map_design_prime
% --------------------

map_design_prime( _, [], [] ).
map_design_prime( ExtentF, [H1|T1], [H2|T2] ) :-
	design_prime( H1, ExtentF, H2 ),
	map_design_prime( ExtentF, T1, T2 ).
   
% fun design_prime
% ----------------

design_prime( node( Label, Subtrees ), 
	      ExtentF,
	      pair( node(pair(Label,0.0), Ptrees), 
		    [Result | MergedList] ) ) :-
  map_design_prime( ExtentF, Subtrees, Mapped ),
  listpair_unzip( Mapped, Trees, Extents ),
  fitlist( Extents, Positions ),
  listpair_map( movetree, Trees, Positions, Ptrees ),
  listpair_map( moveextent, Extents, Positions, Pextents ),
  mergelist( Pextents, MergedList ),
  ToCall =.. [ ExtentF, Label, Result ],
  call( ToCall ).

% fun design
% ----------

design( Tree, ExtentF, Result ) :-
  design_prime( Tree, ExtentF, Result ).

% fun collapse_extent
% -------------------

collapse_extent( [], extent( 0.0, 0.0, 0 ) ) :-
  !.

collapse_extent( [ extent( Left, Right, Height ) | Rest ], 
		 extent( Left_result, Right_result, Height_result ) ) :-
  collapse_extent( Rest, extent( Left_prime, Right_prime, Height_prime ) ),
  Left_result is min( Left, Left_prime ),
  Right_result is max( Right, Right_prime ),
  Height_result is Height + Height_prime.

/* ------------------------------------------------------------- 
   From file: Draw.sml
   ------------------------------------------------------------- */
	
drawline( X1, Y1, X2, Y2 ) :-
  X1int is integer(X1),
  X2int is integer(X2),
  draw_line( X1int, Y1, X2int, Y2 ).

drawlabel( String, X, Y, pair( pair( RetX1, Y ), pair( RetX2, Height ) ) ) :-
  DropAbove = 10,
  DropBelow = 10,
  Xint is integer( X ),
  Yint is integer( Y ),
  draw_label( String, Xint, Yint, Extent ),
  Extent = extent( Left, Right, Height ),
  Sum1 is X + Left,
  RetX1 is integer( Sum1 ),
  Diff1 is Right - Left,
  RetX2 is -(integer(-Diff1)).

pickx( node( pair( Label, X ), Subtrees  ), X ).

eldestx( [H|T], Eldest ) :-
  pickx( H, Eldest ).

youngestx( Subtrees, Youngest ) :-
  reverse( Subtrees, [H|T] ),
  pickx( H, Youngest ).


draw2( Xabs, Y, [extent( Left, Right, Height )|E],
	   node( pair( S, Xrel ), [] ),
	   pair( S, Box ) 
	  ) :-
  X is Xabs + Xrel,
  drawlabel( S, X, Y, Box ),
  !.

draw2( Xabs, 
       Y, 
	   [extent( Left, Right, Height )|E],
	   node( pair( S, Xrel ), Subtrees ), 
	   [ pair( S, Box ) | Tail ]
	  ) :-
  X is Xabs + Xrel,
  drawlabel( S, X, Y, Box ),
  DropBelow = 10,
  Yprime is Y + Height + DropBelow,
  eldestx( Subtrees, R1 ),
  X2 is R1 + X,
  youngestx( Subtrees, R2 ),
  X3 is R2 + X,
  drawline( X2, Yprime, X3, Yprime ),
  Y2 is Y + Height,
  drawline( X, Y2, X, Yprime ),
  do_app( Subtrees, X, Yprime, Y3 ),
  do_draw2_map( X, Y3, E, Subtrees, Draw2 ).
  
  
do_app( [], _, _, _ ) :- !.
do_app( [H|T], X, Yprime, Y3 ) :- 
  pickx( H, PickxSubtrees ),
  Xprime is X + PickxSubtrees,
  DropAbove = 10,
  Y3 is Yprime + DropAbove,
  drawline( Xprime, Yprime, Xprime, Y3 ),
  do_app( T, X, Yprime, Y3 ).
  
do_draw2_map( X, Yplus, E, Subtrees, Out ) :-
  do_draw2_map( X, Yplus, E, Subtrees, Out, [] ).
  
do_draw2_map( X, Yplus, E, [], Out, Out ) :- 
	!.
do_draw2_map( X, Yplus, E, [H|T], Out, SoFar ) :-
  draw2( X, Yplus, E, H, Draw2 ),
  NewSoFar = [Draw2|SoFar],
  do_draw2_map( X, Yplus, E, T, Out, NewSoFar ).
	
/* ------------------------------------------------------------- 
   From file: JavaDraw.sml
   ------------------------------------------------------------- */

append( A, B, C ) :- '$append'( A, B, C ).
reverse( A, B ) :- '$reverse'( A, B ).
max( A, B, C ) :- '$max'( A, B, C ).

get_half_width( S, H ) :- form1( Object ), Object:get_half_width( S, H ).

draw_line( A, B, C, D ) :- form1( Object ), cs_method( Object, 'draw_line'( A, B, C, D ), _ ).
get_drop_below( DB ) :- form1( Object ), cs_method( Object, 'get_drop_below'( ), DB ).
get_drop_above( DB ) :- form1( Object ), cs_method( Object, 'get_drop_above'( ), DB ).
%draw_label( S, X, Y, extent( Left, Right, Height ) ) :- 
draw_label( S, X, Y, extent( -20.0, 20.0, 5 ) ) :-
  form1( Object ), 
  cs_method( Object, 'draw_label'( S, X, Y ), _ ).

end_invokes :- form1( Object ), cs_method( Object, 'end_invokes'( ), _ ).

validate( Class ):-
  form1( Object ), 
  Object:validate( Class, V ), V = 1.
